_base_ = ["./train.py", "../_base_/models/segmentors/seg_class_incr.py", "../_base_/data/custom_seg.py"]

optimizer = dict(_delete_=True, type="Adam", lr=1e-3, eps=1e-08, weight_decay=0.0)

optimizer_config = dict(
    _delete_=True,
    grad_clip=dict(
        # method='adaptive',
        # clip=0.2,
        # method='default',
        max_norm=40,
        norm_type=2,
    ),
)

log_config = dict(
    interval=10,
    hooks=[
        dict(type="TextLoggerHook", by_epoch=True, ignore_last=False),
        # dict(type='TensorboardLoggerHook')
    ],
)

dist_params = dict(backend="nccl", linear_scale_lr=False)

runner = dict(type="EpochRunnerWithCancel", max_epochs=300)

checkpoint_config = dict(by_epoch=True, interval=1)

evaluation = dict(interval=1, metric=["mDice", "mIoU"], show_log=True)

seed = 42
find_unused_parameters = False

task_adapt = dict(
    type="mpa",
    op="REPLACE",
)

ignore = True



"""This module contains mappings from ConfigElementType names to the appropriate constructor classes or functions.

Currently three different mappings are defined: `PrimitiveElementMapping` representing the different base parameter
types, `GroupElementMapping` representing the different configuration groups and `RuleElementMapping` representing the
different ui exposure logic classes
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from enum import Enum
from functools import partial

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import (
    ParameterGroup,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    float_selectable,
    selectable,
)
from otx.api.configuration.ui_rules.rules import Rule, UIRules


class PrimitiveElementMapping(Enum):
    """This Enum represents the mapping from primitive configuration element names to their constructors.

    It is only used by the ConfigHelper to be able to reconstruct a configuration object out of a dictionary
    or yaml string.
    """

    INTEGER = partial(configurable_integer)
    FLOAT = partial(configurable_float)
    BOOLEAN = partial(configurable_boolean)
    FLOAT_SELECTABLE = partial(float_selectable)
    SELECTABLE = partial(selectable)

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name


class GroupElementMapping(Enum):
    """This Enum represents the mapping from configuration group names to their constructors.

    It is only used by the ConfigHelper to be able to reconstruct a configuration object out of a dictionary or yaml
    string.
    """

    PARAMETER_GROUP = ParameterGroup
    CONFIGURABLE_PARAMETERS = ConfigurableParameters

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name


class RuleElementMapping(Enum):
    """This Enum represents the mapping from configuration logic names to their constructors.

    It is only used by the ConfigHelper to be able to reconstruct a configuration object out of a dictionary or yaml
    string.
    """

    RULE = Rule
    UI_RULES = UIRules

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.nn.functional as F
from mmcls.models.builder import HEADS, build_loss
from mmcls.models.heads.base_head import BaseHead
from torch import nn


@HEADS.register_module()
class SupConClsHead(BaseHead):
    """
    Supervised Contrastive Learning head for Classification using SelfSL
    Args:
        num_classes (int): The number of classes of dataset used for training
        in_channels (int): The channels of input data from the backbone
        aux_mlp (dict): A dictionary with the out_channels and optionally the
                         hid_channels of the auxiliary MLP head.
        loss (dict): The SelfSL loss: BarlowTwinsLoss (default)
        topk (set): evaluation topk score, default is (1, )
    """

    def __init__(self, num_classes: int, in_channels: int, aux_mlp, loss, aux_loss, topk=(1,), init_cfg=None, **kwargs):
        if in_channels <= 0:
            raise ValueError(f"in_channels={in_channels} must be a positive integer")
        if num_classes <= 0:
            raise ValueError("at least one class must be exist num_classes.")

        if isinstance(topk, int):
            topk = (topk,)
        for _topk in topk:
            assert _topk > 0, "Top-k should be larger than 0"
        topk = (1,) if num_classes < 5 else (1, 5)
        super().__init__(init_cfg=init_cfg)

        self.topk = topk
        self.compute_loss = build_loss(loss)
        self.aux_loss = build_loss(aux_loss)

        # Set up the standard classification head
        self.num_classes = num_classes
        self.fc = nn.Linear(in_features=in_channels, out_features=self.num_classes)

        # Set up the auxiliar head
        out_channels = aux_mlp["out_channels"]
        if out_channels <= 0:
            raise ValueError(f"out_channels={out_channels} must be a positive integer")
        if "hid_channels" in aux_mlp and aux_mlp["hid_channels"] > 0:
            hid_channels = aux_mlp["hid_channels"]
            self.aux_mlp = nn.Sequential(
                nn.Linear(in_features=in_channels, out_features=hid_channels),
                nn.ReLU(inplace=True),
                nn.Linear(in_features=hid_channels, out_features=out_channels),
            )
        else:
            self.aux_mlp = nn.Linear(in_features=in_channels, out_features=out_channels)

    def forward_train(self, x, gt_label):
        """
        Forward train head using the Supervised Contrastive Loss
        Args:
            x (Tensor): features from the backbone.
        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """

        losses = dict(loss=0.0)
        cls_score = self.fc(x)

        bsz = gt_label.shape[0]
        # make sure we have two views for each label and split them
        assert x.shape[0] == 2 * bsz
        feats1, feats2 = torch.split(self.aux_mlp(x), [bsz, bsz], dim=0)
        gt_label = torch.cat([gt_label, gt_label], dim=0)

        loss = self.compute_loss(cls_score, gt_label)
        aux_loss = self.aux_loss(feats1, feats2)
        losses["loss"] = loss + aux_loss
        return losses

    def simple_test(self, img):
        """
        Test without data augmentation.
        """
        cls_score = self.fc(img)

        if isinstance(cls_score, list):
            cls_score = sum(cls_score) / float(len(cls_score))
        pred = F.softmax(cls_score, dim=1) if cls_score is not None else None
        if torch.onnx.is_in_onnx_export():
            return pred
        pred = list(pred.detach().cpu().numpy())
        return pred



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os.path as osp

from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .inferrer import ClsInferrer

logger = get_logger()


@STAGES.register_module()
class ClsEvaluator(ClsInferrer):
    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run evaluation stage for classification

        - Run inference
        - Run evaluation via MMClassification -> MMCV
        """
        self._init_logger()
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("evaluate!")

        #  # Save config
        #  cfg.dump(osp.join(cfg.work_dir, "config.yaml"))
        #  logger.info(f"Config:\n{cfg.pretty_text}")

        # Inference
        model_builder = kwargs.get("model_builder", None)
        infer_results = super().infer(cfg, model_builder)

        # Evaluate inference results
        eval_cfg = cfg.get("evaluation", {}).copy()
        eval_cfg.pop("by_epoch", None)
        eval_result = self.dataset.evaluate(infer_results, **eval_cfg)

        logger.info(eval_result)
        return eval_result



"""Unit Tests for the MPA Dataset Pipelines OTE Transforms."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from __future__ import annotations

import pytest
import torch
from PIL import Image
from torchvision.transforms import functional as F

from otx.mpa.modules.datasets.pipelines.transforms.ote_transforms import (
    PILToTensor,
    RandomRotate,
    TensorNormalize,
)


@pytest.fixture
def data() -> dict[str, list[str] | Image]:
    # a sample result dictionary to use in tests
    return {
        "img_fields": ["img"],
        "img": Image.new(mode="RGB", size=(224, 224), color="red"),
    }


def test_PILToTensor(data: dict[str, list[str] | Image]) -> None:
    """Test PILToTensor transform."""
    transform = PILToTensor()
    result = transform(data.copy())  # copy to avoid modifying the original data

    assert result["PILToTensor"] is True
    assert result["img_fields"] == ["img"]
    assert torch.equal(result["img"], F.to_tensor(data["img"]))


def test_TensorNormalize(data: dict[str, list[str] | Image]) -> None:
    """Test TensorNormalize transform."""
    mean, std = [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
    data = PILToTensor()(data)  # convert to tensor
    transform = TensorNormalize(mean=mean, std=std)
    result = transform(data.copy())  # copy to avoid modifying the original data

    assert result["TensorNormalize"] is True
    assert result["img_fields"] == ["img"]
    assert torch.equal(result["img"], F.normalize(data["img"], mean, std))


def test_RandomRotate(data: dict[str, list[str] | Image]) -> None:
    """Test RandomRotate transform."""
    transform = RandomRotate(p=1.0, angle=(-10, 10))
    result = transform(data.copy())  # copy to avoid modifying the original data.
    assert result["img"] != data["img"]  # image should be rotated



"""Collection of helper functions for unit tests of otx.algorithms.action."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Dict, List

import numpy as np

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.metadata import MetadataItemEntity, VideoMetadata
from otx.api.entities.model_template import ModelTemplate
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle


class MockImage(Image):
    """Mock class for Image entity."""

    def __init__(self, file_path):
        self.__file_path = file_path
        self.__data = np.ndarray((256, 256, 3))
        super().__init__(self.__data)


class MockPipeline:
    """Mock class for data pipeline.

    It returns its inputs.
    """

    def __call__(self, results: Dict[str, Any]) -> Dict[str, Any]:
        return results


def generate_labels(length: int, domain: Domain) -> List[LabelEntity]:
    """Generate list of LabelEntity given length and domain."""

    output: List[LabelEntity] = []
    for i in range(length):
        output.append(LabelEntity(name=f"{i + 1}", domain=domain, id=ID(i + 1)))
    return output


def generate_action_cls_otx_dataset(video_len: int, frame_len: int, labels: List[LabelEntity]) -> DatasetEntity:
    """Generate otx_dataset for action classification task."""

    items: List[DatasetItemEntity] = []
    for video_id in range(video_len):
        for frame_idx in range(frame_len):
            item = DatasetItemEntity(
                media=MockImage(f"{video_id}_{frame_idx}.png"),
                annotation_scene=AnnotationSceneEntity(
                    annotations=[Annotation(Rectangle.generate_full_box(), [ScoredLabel(labels[video_id])])],
                    kind=AnnotationSceneKind.ANNOTATION,
                ),
                metadata=[MetadataItemEntity(data=VideoMetadata(video_id, frame_idx, is_empty_frame=False))],
            )
            items.append(item)
    dataset = DatasetEntity(items=items)
    return dataset


def generate_action_det_otx_dataset(video_len: int, frame_len: int, labels: List[LabelEntity]) -> DatasetEntity:
    """Generate otx_dataset for action detection task."""

    items: List[DatasetItemEntity] = []
    proposals: Dict[str, List[float]] = {}
    for video_id in range(video_len):
        for frame_idx in range(frame_len):
            if frame_idx % 2 == 0:
                item = DatasetItemEntity(
                    media=MockImage(f"{video_id}_{frame_idx}.png"),
                    annotation_scene=AnnotationSceneEntity(
                        annotations=[Annotation(Rectangle.generate_full_box(), [ScoredLabel(labels[video_id])])],
                        kind=AnnotationSceneKind.ANNOTATION,
                    ),
                    metadata=[MetadataItemEntity(data=VideoMetadata(str(video_id), frame_idx, is_empty_frame=False))],
                )
                proposals[f"{video_id},{frame_idx:04d}"] = [0.0, 0.0, 1.0, 1.0]
            else:
                item = DatasetItemEntity(
                    media=MockImage(f"{video_id}_{frame_idx}.png"),
                    annotation_scene=AnnotationSceneEntity(
                        annotations=[Annotation(Rectangle.generate_full_box(), [ScoredLabel(labels[video_id])])],
                        kind=AnnotationSceneKind.ANNOTATION,
                    ),
                    metadata=[MetadataItemEntity(data=VideoMetadata(str(video_id), frame_idx, is_empty_frame=True))],
                )
            items.append(item)
    dataset = DatasetEntity(items=items)
    return dataset, proposals


class MockModelTemplate(ModelTemplate):
    """Mock class for ModelTemplate."""

    def __post_init__(self):
        pass


def return_args(*args, **kwargs):
    """This function returns its args."""
    return args, kwargs


def return_inputs(inputs):
    """This function returns its input."""
    return inputs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.registry import STAGES
from otx.mpa.seg.inferrer import SegInferrer

from .stage import IncrSegStage


@STAGES.register_module()
class IncrSegInferrer(IncrSegStage, SegInferrer):
    def __init__(self, **kwargs):
        IncrSegStage.__init__(self, **kwargs)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn.functional as F
from mmseg.models.builder import LOSSES
from mmseg.models.losses.utils import get_class_weight

from .mpa_pixel_base import MPABasePixelLoss


@LOSSES.register_module()
class CrossEntropyLossWithIgnore(MPABasePixelLoss):
    """CrossEntropyLossWithIgnore with Ignore Mode Support for Class Incremental Learning.

    Args:
        model_classes (list[str]): Model classes
        bg_aware (bool, optional): Whether to enable BG-aware loss
            'background' class would be added the start of model classes/label schema
        reduction (str, optional): . Defaults to 'mean'.
            Options are "none", "mean" and "sum".
        loss_weight (float, optional): Weight of the loss. Defaults to 1.0.
    """

    def __init__(self, reduction="mean", loss_weight=None, **kwargs):
        super(CrossEntropyLossWithIgnore, self).__init__(**kwargs)

        self.reduction = reduction
        self.class_weight = get_class_weight(loss_weight)

    @property
    def name(self):
        return "ce_with_ignore"

    def _calculate(self, cls_score, label, valid_label_mask, scale):
        if cls_score.shape[0] == 0:
            return torch.tensor(0.0)

        batch_size = label.shape[0]
        label = torch.from_numpy(label).to(cls_score.device)
        probs_all = F.softmax(scale * cls_score, dim=1)
        losses_l = []
        for i in range(batch_size):
            probs_gathered = probs_all[i, valid_label_mask[i] == 1]
            probs_nomatch = probs_all[i, valid_label_mask[i] == 0]
            probs_gathered = torch.unsqueeze(probs_gathered, 0)
            probs_nomatch = torch.unsqueeze(probs_nomatch, 0)

            probs_gathered[:, 0] += probs_nomatch.sum(dim=1)
            each_prob_log = torch.log(probs_gathered)

            # X-entropy: NLL loss w/ log-probabilities & labels
            each_label = torch.unsqueeze(label[i], 0)
            each_label = each_label.to(cls_score.device)
            loss = F.nll_loss(each_prob_log, each_label, reduction="none", ignore_index=self.ignore_index)
            losses_l.append(loss)

        losses = torch.cat(losses_l, dim=0)

        return losses, cls_score



"""API Tests for segmentation training"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os.path as osp
import random
import time
import warnings
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import numpy as np
from bson import ObjectId

from otx.algorithms.common.tasks.training_base import BaseTask
from otx.algorithms.segmentation.tasks import (
    SegmentationInferenceTask,
    SegmentationTrainTask,
)
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.metrics import Performance
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_helpers import generate_random_annotated_image
from tests.test_suite.e2e_test_system import e2e_pytest_api

DEFAULT_SEG_TEMPLATE_DIR = osp.join("otx/algorithms/segmentation/configs", "ocr_lite_hrnet_18_mod2")


def task_eval(task: BaseTask, model: ModelEntity, dataset: DatasetEntity) -> Performance:
    start_time = time.time()
    result_dataset = task.infer(dataset.with_empty_annotations())
    end_time = time.time()
    print(f"{len(dataset)} analysed in {end_time - start_time} seconds")
    result_set = ResultSetEntity(model=model, ground_truth_dataset=dataset, prediction_dataset=result_dataset)
    task.evaluate(result_set)
    assert result_set.performance is not None
    return result_set.performance


class TestMPASegAPI:
    """
    Collection of tests for OTE API and OTE Model Templates
    """

    @e2e_pytest_api
    def test_reading_segmentation_cls_incr_model_template(self):
        segmentation_template = [
            "ocr_lite_hrnet_18_mod2",
            "ocr_lite_hrnet_s_mod2",
            "ocr_lite_hrnet_x_mod3",
        ]
        for model_template in segmentation_template:
            parse_model_template(osp.join("otx/algorithms/segmentation/configs", model_template, "template.yaml"))

    @staticmethod
    def generate_label_schema(label_names):
        label_domain = Domain.SEGMENTATION
        rgb = [int(i) for i in np.random.randint(0, 256, 3)]
        colors = [Color(*rgb) for _ in range(len(label_names))]
        not_empty_labels = [
            LabelEntity(name=name, color=colors[i], domain=label_domain, id=i) for i, name in enumerate(label_names)
        ]
        empty_label = LabelEntity(
            name="Empty label",
            color=Color(42, 43, 46),
            is_empty=True,
            domain=label_domain,
            id=len(not_empty_labels),
        )

        label_schema = LabelSchemaEntity()
        exclusive_group = LabelGroup(name="labels", labels=not_empty_labels, group_type=LabelGroupType.EXCLUSIVE)
        empty_group = LabelGroup(name="empty", labels=[empty_label], group_type=LabelGroupType.EMPTY_LABEL)
        label_schema.add_group(exclusive_group)
        label_schema.add_group(empty_group)
        return label_schema

    def init_environment(self, params, model_template, number_of_images=10):
        labels_names = ("rectangle", "ellipse", "triangle")
        labels_schema = self.generate_label_schema(labels_names)
        labels_list = labels_schema.get_labels(False)
        environment = TaskEnvironment(
            model=None,
            hyper_parameters=params,
            label_schema=labels_schema,
            model_template=model_template,
        )

        warnings.filterwarnings("ignore", message=".* coordinates .* are out of bounds.*")
        items = []
        for i in range(0, number_of_images):
            image_numpy, shapes = generate_random_annotated_image(
                image_width=640,
                image_height=480,
                labels=labels_list,
                max_shapes=20,
                min_size=50,
                max_size=100,
                random_seed=None,
            )
            # Convert all shapes to polygons
            out_shapes = []
            for shape in shapes:
                shape_labels = shape.get_labels(include_empty=True)

                in_shape = shape.shape
                if isinstance(in_shape, Rectangle):
                    points = [
                        Point(in_shape.x1, in_shape.y1),
                        Point(in_shape.x2, in_shape.y1),
                        Point(in_shape.x2, in_shape.y2),
                        Point(in_shape.x1, in_shape.y2),
                    ]
                elif isinstance(in_shape, Ellipse):
                    points = [Point(x, y) for x, y in in_shape.get_evenly_distributed_ellipse_coordinates()]
                elif isinstance(in_shape, Polygon):
                    points = in_shape.points

                out_shapes.append(Annotation(Polygon(points=points), labels=shape_labels))

            image = Image(data=image_numpy)
            annotation = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=out_shapes)
            items.append(DatasetItemEntity(media=image, annotation_scene=annotation))
        warnings.resetwarnings()

        rng = random.Random()
        rng.shuffle(items)
        for i, _ in enumerate(items):
            subset_region = i / number_of_images
            if subset_region >= 0.8:
                subset = Subset.TESTING
            elif subset_region >= 0.6:
                subset = Subset.VALIDATION
            else:
                subset = Subset.TRAINING

            items[i].subset = subset

        dataset = DatasetEntity(items)

        return environment, dataset

    @staticmethod
    def setup_configurable_parameters(template_dir, num_iters=10):
        model_template = parse_model_template(osp.join(template_dir, "template.yaml"))

        hyper_parameters = create(model_template.hyper_parameters.data)
        hyper_parameters.learning_parameters.learning_rate_fixed_iters = 0
        hyper_parameters.learning_parameters.learning_rate_warmup_iters = 1
        hyper_parameters.learning_parameters.num_iters = num_iters
        hyper_parameters.learning_parameters.num_checkpoints = 1

        return hyper_parameters, model_template

    @e2e_pytest_api
    def test_cancel_training_segmentation(self):
        """
        Tests starting and cancelling training.

        Flow of the test:
        - Creates a randomly annotated project with a small dataset.
        - Start training and give cancel training signal after 10 seconds. Assert that training
            stops within 35 seconds after that
        - Start training and give cancel signal immediately. Assert that training stops within 25 seconds.

        This test should be finished in under one minute on a workstation.
        """
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_SEG_TEMPLATE_DIR, num_iters=200)
        segmentation_environment, dataset = self.init_environment(hyper_parameters, model_template, 64)

        segmentation_task = SegmentationTrainTask(task_environment=segmentation_environment)

        executor = ThreadPoolExecutor(max_workers=1, thread_name_prefix="train_thread")

        output_model = ModelEntity(
            dataset,
            segmentation_environment.get_model_configuration(),
        )

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback

        # Test stopping after some time
        start_time = time.time()
        train_future = executor.submit(segmentation_task.train, dataset, output_model, train_parameters)
        # give train_thread some time to initialize the model
        while not segmentation_task._is_training:
            time.sleep(10)
        segmentation_task.cancel_training()

        # stopping process has to happen in less than 35 seconds
        train_future.result()
        assert training_progress_curve[-1] == 100
        assert time.time() - start_time < 100, "Expected to stop within 100 seconds."

        # Test stopping immediately
        start_time = time.time()
        train_future = executor.submit(segmentation_task.train, dataset, output_model)
        segmentation_task.cancel_training()

        train_future.result()
        assert time.time() - start_time < 25  # stopping process has to happen in less than 25 seconds

    @e2e_pytest_api
    def test_training_progress_tracking(self):
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_SEG_TEMPLATE_DIR, num_iters=5)
        segmentation_environment, dataset = self.init_environment(hyper_parameters, model_template, 12)

        task = SegmentationTrainTask(task_environment=segmentation_environment)
        print("Task initialized, model training starts.")

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback
        output_model = ModelEntity(
            dataset,
            segmentation_environment.get_model_configuration(),
        )
        task.train(dataset, output_model, train_parameters)

        assert len(training_progress_curve) > 0
        assert np.all(training_progress_curve[1:] >= training_progress_curve[:-1])

    @e2e_pytest_api
    def test_inference_progress_tracking(self):
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_SEG_TEMPLATE_DIR, num_iters=10)
        segmentation_environment, dataset = self.init_environment(hyper_parameters, model_template, 12)

        task = SegmentationInferenceTask(task_environment=segmentation_environment)
        print("Task initialized, model inference starts.")

        inference_progress_curve = []

        def progress_callback(progress: int):
            assert isinstance(progress, int)
            inference_progress_curve.append(progress)

        inference_parameters = InferenceParameters()
        inference_parameters.update_progress = progress_callback
        task.infer(dataset.with_empty_annotations(), inference_parameters)

        assert len(inference_progress_curve) > 0
        assert np.all(inference_progress_curve[1:] >= inference_progress_curve[:-1])

    @e2e_pytest_api
    def test_inference_task(self):
        # Prepare pretrained weights
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_SEG_TEMPLATE_DIR, num_iters=2)
        segmentation_environment, dataset = self.init_environment(hyper_parameters, model_template, 30)
        val_dataset = dataset.get_subset(Subset.VALIDATION)

        train_task = SegmentationTrainTask(task_environment=segmentation_environment)

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback
        trained_model = ModelEntity(
            dataset,
            segmentation_environment.get_model_configuration(),
        )
        train_task.train(dataset, trained_model, train_parameters)
        performance_after_train = task_eval(train_task, trained_model, val_dataset)

        # Create InferenceTask
        segmentation_environment.model = trained_model
        inference_task = SegmentationInferenceTask(task_environment=segmentation_environment)

        performance_after_load = task_eval(inference_task, trained_model, val_dataset)

        assert performance_after_train == performance_after_load

        # Export
        exported_model = ModelEntity(dataset, segmentation_environment.get_model_configuration(), _id=ObjectId())
        inference_task.export(ExportType.OPENVINO, exported_model)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math

import torch
from mmcv.parallel import is_module_wrapper
from mmcv.runner import HOOKS, Hook
from mmcv.runner.hooks.ema import EMAHook

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class DualModelEMAHook(Hook):
    """Generalized re-implementation of mmcv.runner.EMAHook

    Source model paramters would be exponentially averaged
    onto destination model pararmeters on given intervals

        .. math::

            \text{Xema_{t+1}} = (1 - \text{momentum}) \times
            \text{Xema_{t}} +  \text{momentum} \times X_t

    Args:
        momentum (float): The momentum used for updating ema parameter.
            Defaults to 0.0002.
        epoch_momentum (float): if > 0, momentum is ignored and re-calculated.
            momentum = 1 - exp(1 - epoch_momentum, 1/num_iter_per_epoch)
            Defaults to 0.
        interval (int): Update ema parameter every interval iteration.
            Defaults to 1.
        start_epoch (int): During initial a few epochs, we just copy values
            to update ema parameters. Defaults to 5.
        src_model_name (str): Source model for EMA (X)
        dst_model_name (str): Destination model for EMA (Xema)
    """

    def __init__(
        self,
        momentum=0.0002,
        epoch_momentum=0.0,
        interval=1,
        start_epoch=5,
        src_model_name="model_s",
        dst_model_name="model_t",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.momentum = 1 - (1 - momentum) ** interval
        self.epoch_momentum = epoch_momentum
        self.interval = interval
        self.start_epoch = start_epoch
        self.src_model_name = src_model_name
        self.dst_model_name = dst_model_name
        self.enabled = False

    def before_run(self, runner):
        """Set up src & dst model parameters."""
        model = self._get_model(runner)
        self.src_model = getattr(model, self.src_model_name, None)
        self.dst_model = getattr(model, self.dst_model_name, None)
        if self.src_model and self.dst_model:
            self.enabled = True
            self.src_params = self.src_model.state_dict(keep_vars=True)
            self.dst_params = self.dst_model.state_dict(keep_vars=True)
            if runner.epoch == 0 and runner.iter == 0:
                # If it's not resuming from a checkpoint
                # initialize student model by teacher model
                # (teacher model is main target of load/save)
                # (if it's resuming there will be student weights in checkpoint. No need to copy)
                self._sync_model()
                logger.info("Initialized student model by teacher model")
                logger.info(f"model_s model_t diff: {self._diff_model()}")

    def before_train_epoch(self, runner):
        if self.epoch_momentum > 0.0:
            iter_per_epoch = len(runner.data_loader)
            epoch_decay = 1 - self.epoch_momentum
            iter_decay = math.pow(epoch_decay, self.interval / iter_per_epoch)
            self.momentum = 1 - iter_decay
            logger.info(f"EMA: epoch_decay={epoch_decay} / iter_decay={iter_decay}")
            self.epoch_momentum = 0.0  # disable re-compute

    def after_train_iter(self, runner):
        """Update ema parameter every self.interval iterations."""
        if not self.enabled:
            return

        if runner.iter % self.interval != 0:
            # Skip update
            return

        if runner.epoch + 1 < self.start_epoch:
            # Just copy parameters before start epoch
            self._copy_model()
            return

        # EMA
        self._ema_model()

    def after_train_epoch(self, runner):
        if self.enabled:
            logger.info(f"model_s model_t diff: {self._diff_model()}")

    def _get_model(self, runner):
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        return model

    def _sync_model(self):
        with torch.no_grad():
            for name, src_param in self.src_params.items():
                dst_param = self.dst_params[name]
                src_param.data.copy_(dst_param.data)

    def _copy_model(self):
        with torch.no_grad():
            for name, src_param in self.src_params.items():
                dst_param = self.dst_params[name]
                dst_param.data.copy_(src_param.data)

    def _ema_model(self):
        momentum = min(self.momentum, 1.0)
        with torch.no_grad():
            for name, src_param in self.src_params.items():
                dst_param = self.dst_params[name]
                # dst_param.data.mul_(1 - momentum).add_(src_param.data, alpha=momentum)
                dst_param.data.copy_(dst_param.data * (1 - momentum) + src_param.data * momentum)

    def _diff_model(self):
        diff_sum = 0.0
        with torch.no_grad():
            for name, src_param in self.src_params.items():
                dst_param = self.dst_params[name]
                diff = ((src_param - dst_param) ** 2).sum()
                diff_sum += diff
        return diff_sum


@HOOKS.register_module()
class CustomModelEMAHook(EMAHook):
    def __init__(self, momentum=0.0002, epoch_momentum=0.0, interval=1, **kwargs):
        super().__init__(momentum=momentum, interval=interval, **kwargs)
        self.momentum = momentum
        self.epoch_momentum = epoch_momentum
        self.interval = interval

    def before_train_epoch(self, runner):
        if self.epoch_momentum > 0.0:
            iter_per_epoch = len(runner.data_loader)
            epoch_decay = 1 - self.epoch_momentum
            iter_decay = math.pow(epoch_decay, self.interval / iter_per_epoch)
            self.momentum = 1 - iter_decay
            logger.info(f"Update EMA momentum: {self.momentum}")
            self.epoch_momentum = 0.0  # disable re-compute

        super().before_train_epoch(runner)



"""Visualizer for results of prediction."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from typing import Optional

import cv2
import numpy as np

from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.utils.shape_drawer import ShapeDrawer


class IVisualizer(metaclass=abc.ABCMeta):
    """Interface for converter."""

    @abc.abstractmethod
    def draw(
        self,
        image: np.ndarray,
        annotation: AnnotationSceneEntity,
        meta: dict,
    ) -> np.ndarray:
        """Draw annotations on the image.

        Args:
            image: Input image
            annotation: Annotations to be drawn on the input image
            metadata: Metadata is needed to render

        Returns:
            Output image with annotations.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def show(self, image: np.ndarray) -> None:
        """Show result image."""

        raise NotImplementedError

    @abc.abstractmethod
    def is_quit(self) -> bool:
        """Check if user wishes to quit."""

        raise NotImplementedError


class Visualizer(IVisualizer):
    """Visualize the predicted output by drawing the annotations on the input image.

    Example:
        >>> predictions = inference_model.predict(frame)
        >>> annotation = prediction_converter.convert_to_annotation(predictions)
        >>> output = visualizer.draw(frame, annotation.shape, annotation.get_labels())
        >>> visualizer.show(output)
    """

    def __init__(
        self,
        window_name: Optional[str] = None,
        show_count: bool = False,
        is_one_label: bool = False,
        no_show: bool = False,
        delay: Optional[int] = None,
    ) -> None:
        self.window_name = "Window" if window_name is None else window_name
        self.shape_drawer = ShapeDrawer(show_count, is_one_label)

        self.delay = delay
        self.no_show = no_show
        if delay is None:
            self.delay = 1

    def draw(
        self,
        image: np.ndarray,
        annotation: AnnotationSceneEntity,
        meta: Optional[dict] = None,
    ) -> np.ndarray:
        """Draw annotations on the image.

        Args:
            image: Input image
            annotation: Annotations to be drawn on the input image

        Returns:
            Output image with annotations.
        """

        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

        return self.shape_drawer.draw(image, annotation, labels=[])

    def show(self, image: np.ndarray) -> None:
        """Show result image.

        Args:
            image (np.ndarray): Image to be shown.
        """

        if not self.no_show:
            cv2.imshow(self.window_name, image)

    def is_quit(self) -> bool:
        """Check user wish to quit."""
        if self.no_show:
            return False

        return ord("q") == cv2.waitKey(self.delay)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field

import torch

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class MultiplyV1Attribute(Attribute):
    auto_broadcast: str = field(default="numpy")


@OPS.register()
class MultiplyV1(Operation[MultiplyV1Attribute]):
    TYPE = "Multiply"
    VERSION = 1
    ATTRIBUTE_FACTORY = MultiplyV1Attribute

    def forward(self, input_0, input_1):
        broadcast = self.attrs.auto_broadcast

        if broadcast == "none":
            assert input_0.shape == input_1.shape
            return input_0 * input_1
        elif broadcast == "numpy":
            return input_0 * input_1
        else:
            raise NotImplementedError


@dataclass
class DivideV1Attribute(Attribute):
    m_pythondiv: bool = field(default=True)
    auto_broadcast: str = field(default="numpy")


@OPS.register()
class DivideV1(Operation[DivideV1Attribute]):
    TYPE = "Divide"
    VERSION = 1
    ATTRIBUTE_FACTORY = DivideV1Attribute

    def forward(self, input_0, input_1):
        broadcast = self.attrs.auto_broadcast

        if broadcast == "none":
            assert input_0.shape == input_1.shape
            output = input_0 / input_1
        elif broadcast == "numpy":
            output = input_0 / input_1
        else:
            raise NotImplementedError

        non_integer_types = [torch.float16, torch.float32, torch.float64, torch.bool]
        if self.attrs.m_pythondiv and input_0.dtype not in non_integer_types and input_1.dtype not in non_integer_types:
            output = output.type(input_0.dtype)

        return output


@dataclass
class AddV1Attribute(Attribute):
    auto_broadcast: str = field(default="numpy")


@OPS.register()
class AddV1(Operation[AddV1Attribute]):
    TYPE = "Add"
    VERSION = 1
    ATTRIBUTE_FACTORY = AddV1Attribute

    def forward(self, input_0, input_1):
        broadcast = self.attrs.auto_broadcast

        if broadcast == "none":
            assert input_0.shape == input_1.shape
            return input_0 + input_1
        elif broadcast == "numpy":
            return input_0 + input_1
        else:
            raise NotImplementedError


@dataclass
class SubtractV1Attribute(Attribute):
    auto_broadcast: str = field(default="numpy")


@OPS.register()
class SubtractV1(Operation[SubtractV1Attribute]):
    TYPE = "Subtract"
    VERSION = 1
    ATTRIBUTE_FACTORY = SubtractV1Attribute

    def forward(self, input_0, input_1):
        broadcast = self.attrs.auto_broadcast

        if broadcast == "none":
            assert input_0.shape == input_1.shape
            return input_0 - input_1
        elif broadcast == "numpy":
            return input_0 - input_1
        else:
            raise NotImplementedError


@dataclass
class TanV0Attribute(Attribute):
    pass


@OPS.register()
class TanV0(Operation[TanV0Attribute]):
    TYPE = "Tan"
    VERSION = 0
    ATTRIBUTE_FACTORY = TanV0Attribute

    def forward(self, input):
        return torch.tan(input)



_base_ = ["./train.py", "../_base_/data/coco_ubt.py", "../_base_/models/detectors/detector.py"]

model = dict(super_type="UnbiasedTeacher")  # Used as general framework

custom_hooks = [
    dict(
        type="DualModelEMAHook",
        epoch_momentum=0.4,
        start_epoch=2,
    ),
    dict(
        type="LazyEarlyStoppingHook",
        start=3,
        patience=10,
        iteration_patience=0,
        metric="bbox_mAP",
        interval=1,
        priority=75,
    ),
]



_base_ = ["./pipelines/incr_seg.py"]

__dataset_type = ""
__data_root = ""
__train_pipeline = {{_base_.train_pipeline}}
__test_pipeline = {{_base_.test_pipeline}}

data = dict(
    samples_per_gpu=4,
    workers_per_gpu=2,
    train=dict(
        type="RepeatDataset",
        times=1,
        dataset=dict(
            type=__dataset_type,
            data_root=__data_root,
            img_dir=None,
            ann_dir=None,
            split=None,
            pipeline=__train_pipeline,
        ),
    ),
    val=dict(
        type=__dataset_type, data_root=__data_root, img_dir=None, ann_dir=None, split=None, pipeline=__test_pipeline
    ),
    test=dict(
        type=__dataset_type, data_root=__data_root, img_dir=None, ann_dir=None, split=None, pipeline=__test_pipeline
    ),
)



"""Adapt AVADataset in mmaction2 into OTXDataset."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from collections import defaultdict
from copy import copy, deepcopy
from logging import Logger
from typing import Any, Dict, List, Sequence, Tuple

import mmcv
import numpy as np
from mmaction.core.evaluation.ava_utils import det2csv
from mmaction.datasets.ava_dataset import AVADataset
from mmaction.datasets.builder import DATASETS
from mmaction.datasets.pipelines import Compose
from mmaction.utils import get_root_logger
from mmcv.utils import print_log

from otx.algorithms.action.adapters.mmaction.data.pipelines import RawFrameDecode
from otx.algorithms.action.adapters.mmaction.utils import det_eval
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.metadata import VideoMetadata
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.shape_factory import ShapeFactory

root_logger = get_root_logger()


# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-locals, super-init-not-called
@DATASETS.register_module()
class OTXActionDetDataset(AVADataset):
    """Wrapper that allows using a OTX dataset to train action detection models.

    This wrapper is not based on the filesystem,
    but instead loads the items here directly from the OTX DatasetEntity object.
    It is adapted from AVADataset of mmaction, but it supports other dataset such as UCF and JHMDB.
    """

    class _DataInfoProxy:
        def __init__(
            self,
            otx_dataset: DatasetEntity,
            labels: List[LabelEntity],
            person_det_score_thr: float = 0.9,
            num_max_proposals: int = 1000,
            modality: str = "RGB",
            fps: int = 30,
        ):
            self.otx_dataset = deepcopy(otx_dataset)
            self.labels = labels
            self.label_idx = {label.id: i for i, label in enumerate(labels)}
            self.person_det_score_thr = person_det_score_thr
            self.num_max_proposals = num_max_proposals
            self.modality = modality
            self.fps = fps
            self.data_root = "/" + os.path.join(*os.path.abspath(str(self.otx_dataset[0].media.path)).split("/")[:-4])
            self.proposal_file_name = os.path.abspath(str(self.otx_dataset[0].media.path)).split("/")[-4]
            self.proposal_file = os.path.join(self.data_root, f"{self.proposal_file_name}.pkl")
            self.video_info: Dict[str, Any] = {}

            if os.path.exists(self.proposal_file):
                self.proposals = mmcv.load(self.proposal_file)
                self._patch_proposals()
            else:
                self.proposals = None

            self._update_meta_data()

        def __len__(self) -> int:
            return len(self.otx_dataset)

        def _update_meta_data(self):
            """Update video metadata of each item in self.otx_dataset.

            During iterating DatasetEntity, this function generate video_info(dictionary) to record metadata of video
            After that, this function update metadata of each DatasetItemEntity of DatasetEntity
                - start_index: Offset for the video, this value will be added to sampled frame indices
                - timestamp: Timestamp of the DatasetItemEntity
                - timestamp_start: Start timestamp of the video, this will be used to generate shot_info
                - timestamp_end: End timestamp of the video, this will be used to generate shot_info
                - shot_info = (0, (timestamp_end - timestamp_start)) * self.fps:
                              Range of frame indices, this is used to sample frame indices
                - img_key = "video_id,frame_idx": key of pre-proposal dictionary
                - modality = Modality of data, 'RGB' or 'Flow(Optical Flow)'
            This function removes empty frames(frame with no action), since they are only used for background of clips
            """

            video_info = {}
            start_index = 0
            for idx, item in enumerate(self.otx_dataset):
                metadata = item.get_metadata()[0].data
                if metadata.video_id in video_info:
                    video_info[metadata.video_id]["start_index"] = start_index
                    if metadata.frame_idx < video_info[metadata.video_id]["timestamp_start"]:
                        video_info[metadata.video_id]["timestamp_start"] = metadata.frame_idx
                    if metadata.frame_idx > video_info[metadata.video_id]["timestamp_end"]:
                        video_info[metadata.video_id]["timestamp_end"] = metadata.frame_idx
                else:
                    video_info[metadata.video_id] = {
                        "start_index": idx,
                        "timestamp_start": metadata.frame_idx,
                        "timestamp_end": metadata.frame_idx,
                    }
                    start_index = idx

            remove_indices = []
            for idx, item in enumerate(self.otx_dataset):
                metadata = item.get_metadata()[0].data
                if metadata.is_empty_frame:
                    remove_indices.append(idx)
                    continue

                for key, value in video_info[metadata.video_id].items():
                    metadata.update(key, value)

                shot_info = (0, (metadata.timestamp_end - metadata.timestamp_start)) * self.fps
                img_key = f"{metadata.video_id},{metadata.frame_idx}"
                ignored_labels = np.array([self.label_idx[label.id] for label in item.ignored_labels])
                metadata.update("shot_info", shot_info)
                metadata.update("img_key", img_key)
                metadata.update("timestamp", metadata.frame_idx)
                metadata.update("ignored_labels", ignored_labels)
                metadata.update("modality", self.modality)

                anns = item.get_annotations()
                self._update_annotations(metadata, anns)

            self.otx_dataset.remove_at_indices(remove_indices)
            self.video_info.update(video_info)

        def __getitem__(self, index: int) -> Dict[str, Any]:
            """Prepare a dict 'data_info' that is expected by the mmaction pipeline to handle images and annotations.

            :return data_info: dictionary that contains the image and image metadata, as well as the labels of
            the objects in the image
            This iterates self.otx_dataset, which removes empty frames
            """

            item = self.otx_dataset[index]
            metadata = item.get_metadata()[0].data

            data_info = dict(
                **metadata.metadata,
                ann_info=dict(label_list=self.labels),
                fps=self.fps,
            )

            return data_info

        def _patch_proposals(self):
            """Remove fixed string format.

            AVA dataset pre-proposals have fixed string format.
            Fixed string format have scalability issues so here we remove it
            """
            # FIXME This may consume lots of time depends on size of proposals
            root_logger.info("Patching pre proposals...")
            for img_key in list(self.proposals):
                proposal = self.proposals.pop(img_key)
                new_img_key = img_key.split(",")[0] + "," + str(int(img_key.split(",")[1]))
                self.proposals[f"{new_img_key}"] = proposal
            root_logger.info("Done.")

        def _update_annotations(self, metadata: VideoMetadata, anns: List[Annotation]):
            """Update annotation information to item's metadata."""
            if len(anns) > 0:
                bboxes, labels = [], []
                for ann in anns:
                    rectangle = ShapeFactory.shape_as_rectangle(ann.shape)
                    bbox = np.asarray([rectangle.x1, rectangle.y1, rectangle.x2, rectangle.y2])
                    valid_labels = np.array([int(label.id) for label in ann.get_labels()], dtype=int)
                    label = np.zeros(len(self.labels) + 1, dtype=np.float32)
                    label[valid_labels] = 1.0
                    bboxes.append(bbox)
                    labels.append(label)
                metadata.update("gt_bboxes", np.stack(bboxes))
                metadata.update("gt_labels", np.stack(labels))
            else:
                # Insert dummy gt bboxes for data pipeline in mmaction
                metadata.update("gt_bboxes", np.zeros((1, 4)))

            if self.proposals is not None:
                if metadata.img_key in self.proposals:  # type: ignore[attr-defined]
                    proposals = self.proposals[metadata.img_key]  # type: ignore[attr-defined]
                else:
                    proposals = np.array([[0, 0, 1, 1, 1]])
                thr = min(self.person_det_score_thr, max(proposals[:, 4]))
                positive_inds = proposals[:, 4] >= thr
                proposals = proposals[positive_inds]
                proposals = proposals[: self.num_max_proposals]
                metadata.update("proposals", proposals[:, :4])
                metadata.update("scores", proposals[:, 4])

    @check_input_parameters_type({"otx_dataset": DatasetParamTypeCheck})
    # TODO Remove duplicated codes with mmaction's AVADataset
    def __init__(
        self,
        otx_dataset: DatasetEntity,
        labels: List[LabelEntity],
        pipeline: Sequence[dict],
        test_mode: bool = False,
        person_det_score_thr: float = 0.9,
        num_max_proposals: int = 1000,
        modality: str = "RGB",
        fps: int = 30,
    ):
        self.otx_dataset = otx_dataset
        self.labels = labels
        self.test_mode = test_mode
        self.modality = modality
        self._FPS = fps
        self.person_det_score_thr = person_det_score_thr
        self.num_max_proposals = num_max_proposals

        # OTX does not support custom_classes
        self.custom_classes = None

        self.video_infos = OTXActionDetDataset._DataInfoProxy(
            otx_dataset, labels, person_det_score_thr, num_max_proposals, modality, fps
        )

        self.pipeline = Compose(pipeline)
        for transform in self.pipeline.transforms:
            if isinstance(transform, RawFrameDecode):
                transform.otx_dataset = self.otx_dataset

        # TODO. Handle exclude file for AVA dataset
        self.exclude_file = None

    @check_input_parameters_type()
    def prepare_train_frames(self, idx: int) -> Dict[str, Any]:
        """Get training data and annotations after pipeline.

        Args:
            idx (int): Index of data
        """
        item = copy(self.video_infos[idx])  # Copying dict(), not contents
        return self.pipeline(item)

    @check_input_parameters_type()
    def prepare_test_frames(self, idx: int) -> Dict[str, Any]:
        """Get testing data after pipeline.

        Args:
            idx (int): Index of data
        """
        item = copy(self.video_infos[idx])  # Copying dict(), not contents
        return self.pipeline(item)

    # pylint: disable=too-many-locals, unused-argument
    def evaluate(
        self,
        results: List[List[np.ndarray]],
        metrics: Tuple[str] = ("mAP",),
        logger: Logger = None,
        **kwargs,
    ):
        """Evaluate the prediction results and report mAP."""
        assert len(metrics) == 1 and metrics[0] == "mAP", (
            'For evaluation on AVADataset, you need to use metrics "mAP" '
            "See https://github.com/open-mmlab/mmaction2/pull/567 "
            "for more info."
        )
        csv_results = det2csv(self, results, self.custom_classes)
        predictions = self._get_predictions(csv_results)

        ret = {}
        for metric in metrics:
            msg = f"Evaluating {metric} ..."
            if logger is None:
                msg = "\n" + msg
            print_log(msg, logger=logger)

            eval_result = det_eval(
                predictions,
                metric,
                self.labels,
                self.video_infos,
                self.exclude_file,
                custom_classes=self.custom_classes,
            )
            log_msg = []
            for key, value in eval_result.items():
                log_msg.append(f"\n{key}\t{value: .4f}")
            str_log_msg = "".join(log_msg)
            print_log(str_log_msg, logger=logger)
            ret.update(eval_result)
        return ret

    @staticmethod
    def _get_predictions(csv_results: List[Tuple]):
        """Convert model's inference results to predictions."""
        np_csv_results = np.array(csv_results)
        _img_keys = np_csv_results[:, :2]
        _boxes = np_csv_results[:, 2:6]
        _labels = np_csv_results[:, 6]
        _scores = np_csv_results[:, 7]

        boxes = defaultdict(list)
        labels = defaultdict(list)
        scores = defaultdict(list)

        for _img_key, _box, _label, _score in zip(_img_keys, _boxes, _labels, _scores):
            img_key = _img_key[0] + "," + _img_key[1]
            box = _box.astype("float")
            label = _label.astype("int")
            score = _score.astype("float")
            boxes[img_key].append(box)
            labels[img_key].append(label)
            scores[img_key].append(score)
        return (boxes, labels, scores)



"""Configuration Enums.

Enums needed to define the options of selectable parameters in the configurable
parameter classes.
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from otx.api.configuration import ConfigurableEnum


class POTQuantizationPreset(ConfigurableEnum):
    """POT Quantization Preset Enum.

    This Enum represents the quantization preset for post training optimization.
    """

    PERFORMANCE = "Performance"
    MIXED = "Mixed"


class EarlyStoppingMetrics(ConfigurableEnum):
    """Early Stopping Metric Enum.

    This enum represents the different metrics that can be used for early
    stopping.
    """

    IMAGE_ROC_AUC = "image_AUROC"
    IMAGE_F1 = "image_F1Score"


class ModelName(ConfigurableEnum):
    """Model Name Enum.

    This enum represents the different model architectures for anomaly
    classification.
    """

    STFPM = "stfpm"
    PADIM = "padim"


class ModelBackbone(ConfigurableEnum):
    """Model Backbone Enum.

    This enum represents the common backbones that can be used with Padim and
    STFPM.
    """

    RESNET18 = "resnet18"
    WIDE_RESNET_50 = "wide_resnet50_2"



"""Model Configuration of SSD model for Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = [
    "../../../../common/adapters/mmcv/configs/backbones/mobilenet_v2_w1.yaml",
    "../../base/models/single_stage_detector.py",
]

__width_mult = 1.0

model = dict(
    type="CustomSingleStageDetector",
    bbox_head=dict(
        type="CustomSSDHead",
        num_classes=80,
        in_channels=(int(__width_mult * 96), int(__width_mult * 320)),
        use_depthwise=True,
        norm_cfg=dict(type="BN"),
        act_cfg=dict(type="ReLU"),
        init_cfg=dict(type="Xavier", layer="Conv2d", distribution="uniform"),
        loss_balancing=False,
        anchor_generator=dict(
            type="SSDAnchorGeneratorClustered",
            strides=(16, 32),
            reclustering_anchors=True,
            widths=[
                [
                    38.641007923271076,
                    92.49516032784699,
                    271.4234764938237,
                    141.53469410876247,
                ],
                [
                    206.04136086566515,
                    386.6542727907841,
                    716.9892752215089,
                    453.75609561761405,
                    788.4629155558277,
                ],
            ],
            heights=[
                [
                    48.9243877087132,
                    147.73088476194903,
                    158.23569788707474,
                    324.14510379107367,
                ],
                [
                    587.6216059488938,
                    381.60024152086544,
                    323.5988913027747,
                    702.7486097568518,
                    741.4865860938451,
                ],
            ],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=(0.0, 0.0, 0.0, 0.0),
            target_stds=(0.1, 0.1, 0.2, 0.2),
        ),
    ),
    train_cfg=dict(
        assigner=dict(
            pos_iou_thr=0.4,
            neg_iou_thr=0.4,
        ),
        use_giou=False,
        use_focal=False,
    ),
    backbone=dict(
        out_indices=(
            4,
            5,
        )
    ),
)

load_from = "https://storage.openvinotoolkit.org/repositories/openvino_training_extensions\
/models/object_detection/v2/mobilenet_v2-2s_ssd-992x736.pth"

fp16 = dict(loss_scale=512.0)
ignore = False



# Copyright (c) 2020-2021 The MMSegmentation Authors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from .aggregator import IterativeAggregator, IterativeConcatAggregator
from .angular_pw_conv import AngularPWConv
from .asymmetric_position_attention import AsymmetricPositionAttentionModule
from .channel_shuffle import channel_shuffle
from .local_attention import LocalAttentionModule
from .normalize import normalize
from .psp_layer import PSPModule

__all__ = [
    "IterativeAggregator",
    "IterativeConcatAggregator",
    "channel_shuffle",
    "LocalAttentionModule",
    "PSPModule",
    "AsymmetricPositionAttentionModule",
    "AngularPWConv",
    "normalize",
]



"""OTX Adapters - mmdet.utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .builder import build_classifier
from .config_utils import (
    patch_config,
    patch_datasets,
    patch_evaluation,
    prepare_for_training,
)

__all__ = [
    "patch_config",
    "patch_datasets",
    "patch_evaluation",
    "prepare_for_training",
    "build_classifier",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.yolo_head import YOLOV3Head

from ...mmov_model import MMOVModel


@HEADS.register_module()
class MMOVYOLOV3Head(YOLOV3Head):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
        *args,
        **kwargs,
    ):
        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._inputs = deepcopy(inputs)
        self._outputs = deepcopy(outputs)
        self._init_weight = init_weight
        self._verify_shape = verify_shape

        # dummy input
        in_channels = (512, 256, 128)
        out_channels = (1024, 512, 256)
        if "featmap_strides" in kwargs:
            in_channels = kwargs["featmap_strides"]
            out_channels = kwargs["featmap_strides"]
        super().__init__(in_channels=in_channels, out_channels=out_channels, *args, **kwargs)

    def _init_layers(self):
        self.convs_bridge = torch.nn.ModuleList()
        self.convs_pred = torch.nn.ModuleList()

        for inputs, outputs in zip(self._inputs["convs_bridge"], self._outputs["convs_bridge"]):
            self.convs_bridge.append(
                MMOVModel(
                    self._model_path_or_model,
                    self._weight_path,
                    inputs=inputs,
                    outputs=outputs,
                    remove_normalize=False,
                    init_weight=self._init_weight,
                    verify_shape=self._verify_shape,
                )
            )

        for inputs, outputs in zip(self._inputs["convs_pred"], self._outputs["convs_pred"]):
            self.convs_pred.append(
                MMOVModel(
                    self._model_path_or_model,
                    self._weight_path,
                    inputs=inputs,
                    outputs=outputs,
                    remove_normalize=False,
                    init_weight=self._init_weight,
                    verify_shape=self._verify_shape,
                )
            )

    def init_weights(self):
        # TODO
        pass



"""MobileNet-V3-large-1 for multi-label config."""

# pylint: disable=invalid-name

_base_ = "../base/models/mobilenet_v3.py"

model = dict(
    type="SAMImageClassifier",
    task="classification",
    backbone=dict(mode="large"),
    head=dict(
        type="CustomMultiLabelNonLinearClsHead",
        in_channels=960,
        hid_channels=1280,
        normalized=True,
        scale=7.0,
        act_cfg=dict(
            type="PReLU",
        ),
        loss=dict(
            type="AsymmetricAngularLossWithIgnore",
            gamma_pos=0.0,
            gamma_neg=1.0,
            reduction="sum",
        ),
    ),
)

fp16 = dict(loss_scale=512.0)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field

import torch

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class MatMulV0Attribute(Attribute):
    transpose_a: bool = field(default=False)
    transpose_b: bool = field(default=False)


@OPS.register()
class MatMulV0(Operation[MatMulV0Attribute]):
    TYPE = "MatMul"
    VERSION = 0
    ATTRIBUTE_FACTORY = MatMulV0Attribute

    def forward(self, input_a, input_b):
        if self.attrs.transpose_a:
            input_a = torch.transpose(input_a, -1, -2)
        if self.attrs.transpose_b:
            input_b = torch.transpose(input_b, -1, -2)
        return torch.matmul(input_a, input_b)


@dataclass
class EinsumV7Attribute(Attribute):
    equation: str


@OPS.register()
class EinsumV7(Operation[EinsumV7Attribute]):
    TYPE = "Einsum"
    VERSION = 7
    ATTRIBUTE_FACTORY = EinsumV7Attribute

    def forward(self, *inputs):
        return torch.einsum(self.attrs.equation, *inputs)



"""MobileNet-V3-large-075 for self-supervised learning config."""

# pylint: disable=invalid-name

_base_ = "../model.py"

model = dict(
    type="BYOL",
    task="classification",
    base_momentum=0.996,
    neck=dict(type="SelfSLMLP", in_channels=720, hid_channels=4096, out_channels=256, with_avg_pool=True),
    head=dict(
        type="ConstrastiveHead",
        predictor=dict(type="SelfSLMLP", in_channels=256, hid_channels=4096, out_channels=256, with_avg_pool=False),
    ),
)

load_from = None

resume_from = None

fp16 = None



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import os

import numpy as np
import pytest
from openvino.model_zoo.model_api.models import Model

import otx.algorithms.detection.tasks.openvino
from otx.algorithms.detection.configs.base import DetectionConfig
from otx.algorithms.detection.tasks.openvino import (
    OpenVINODetectionInferencer,
    OpenVINODetectionTask,
    OpenVINOMaskInferencer,
    OpenVINORotatedRectInferencer,
)
from otx.algorithms.detection.utils import generate_label_schema
from otx.api.configuration.helper import create
from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.metrics import Performance, ScoreMetric
from otx.api.entities.model_template import (
    TaskType,
    parse_model_template,
    task_type_to_label_domain,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    DEFAULT_DET_TEMPLATE_DIR,
    DEFAULT_ISEG_TEMPLATE_DIR,
    generate_det_dataset,
    init_environment,
)


class TestOpenVINODetectionInferencer:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        classes = ("rectangle", "ellipse", "triangle")
        self.ov_inferencer = dict()
        task_type = TaskType.DETECTION
        model_template = parse_model_template(os.path.join(DEFAULT_DET_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        params = DetectionConfig(header=hyper_parameters.header)
        label_schema = generate_label_schema(classes, task_type_to_label_domain(task_type))
        mocker.patch("otx.algorithms.detection.tasks.openvino.OpenvinoAdapter")
        mocked_model = mocker.patch.object(Model, "create_model")
        mocked_model.return_value = mocker.MagicMock(spec=Model)
        self.ov_inferencer = OpenVINODetectionInferencer(params, label_schema, "")
        self.fake_input = np.full((5, 1), 0.1)

    @e2e_pytest_unit
    def test_pre_process(self):
        """Test pre_process method in OpenVINODetectionInferencer."""
        self.ov_inferencer.model.preprocess.return_value = {"foo": "bar"}
        returned_value = self.ov_inferencer.pre_process(self.fake_input)
        assert returned_value == {"foo": "bar"}

    @e2e_pytest_unit
    def test_post_process(self):
        """Test post_process method in OpenVINODetectionInferencer."""
        fake_prediction = {
            "boxes": np.random.rand(1, 5, 5),
            "labels": np.array([[0, 0, 0, 1, 0]]),
            "feature_vector": np.random.rand(1, 320, 1, 1),
            "saliency_map": np.random.rand(1, 2, 6, 8),
        }
        fake_metadata = {"original_shape": (480, 640, 3), "resized_shape": (736, 992, 3)}
        returned_value = self.ov_inferencer.post_process(fake_prediction, fake_metadata)
        assert isinstance(returned_value, AnnotationSceneEntity)

    @e2e_pytest_unit
    def test_predict(self, mocker):
        """Test predict method in OpenVINODetectionInferencer."""
        fake_output = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=[])
        mock_pre_process = mocker.patch.object(OpenVINODetectionInferencer, "pre_process", return_value=("", ""))
        mock_forward = mocker.patch.object(OpenVINODetectionInferencer, "forward")
        mock_post_process = mocker.patch.object(OpenVINODetectionInferencer, "post_process", return_value=fake_output)
        returned_value, _ = self.ov_inferencer.predict(self.fake_input)

        mock_pre_process.assert_called_once()
        mock_forward.assert_called_once()
        mock_post_process.assert_called_once()
        assert returned_value == fake_output

    @e2e_pytest_unit
    def test_forward(self):
        """Test forward method in OpenVINODetectionInferencer."""
        fake_output = {"pred": np.full((5, 1), 0.9)}
        self.ov_inferencer.model.infer_sync.return_value = fake_output
        returned_value = self.ov_inferencer.forward({"image": self.fake_input})
        assert returned_value == fake_output


class TestOpenVINOMaskInferencer:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        classes = ("rectangle", "ellipse", "triangle")
        self.ov_inferencer = dict()
        task_type = TaskType.INSTANCE_SEGMENTATION
        model_template = parse_model_template(os.path.join(DEFAULT_ISEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        params = DetectionConfig(header=hyper_parameters.header)
        label_schema = generate_label_schema(classes, task_type_to_label_domain(task_type))
        mocker.patch("otx.algorithms.detection.tasks.openvino.OpenvinoAdapter")
        mocked_model = mocker.patch.object(Model, "create_model")
        mocked_model.return_value = mocker.MagicMock(spec=Model)
        self.ov_inferencer = OpenVINOMaskInferencer(params, label_schema, "")
        self.fake_input = np.full((5, 1), 0.1)

    @e2e_pytest_unit
    def test_pre_process(self):
        """Test pre_process method in OpenVINOMaskInferencer."""
        self.ov_inferencer.model.preprocess.return_value = {"foo": "bar"}
        returned_value = self.ov_inferencer.pre_process(self.fake_input)
        assert returned_value == {"foo": "bar"}


class TestOpenVINORotatedRectInferencer:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        classes = ("rectangle", "ellipse", "triangle")
        self.ov_inferencer = dict()
        task_type = TaskType.DETECTION
        model_template = parse_model_template(os.path.join(DEFAULT_DET_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        params = DetectionConfig(header=hyper_parameters.header)
        label_schema = generate_label_schema(classes, task_type_to_label_domain(task_type))
        mocker.patch("otx.algorithms.detection.tasks.openvino.OpenvinoAdapter")
        mocked_model = mocker.patch.object(Model, "create_model")
        mocked_model.return_value = mocker.MagicMock(spec=Model)
        self.ov_inferencer = OpenVINORotatedRectInferencer(params, label_schema, "")
        self.fake_input = np.full((5, 1), 0.1)

    @e2e_pytest_unit
    def test_pre_process(self):
        """Test pre_process method in RotatedRectInferencer."""
        self.ov_inferencer.model.preprocess.return_value = {"foo": "bar"}
        returned_value = self.ov_inferencer.pre_process(self.fake_input)
        assert returned_value == {"foo": "bar"}


class TestOpenVINODetectionTask:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, otx_model) -> None:

        classes = ("rectangle", "ellipse", "triangle")
        self.ov_inferencer = dict()
        task_type = TaskType.DETECTION

        model_template = parse_model_template(os.path.join(DEFAULT_DET_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        label_schema = generate_label_schema(classes, task_type_to_label_domain(task_type))
        task_env = init_environment(hyper_parameters, model_template, task_type=task_type)
        params = DetectionConfig(header=hyper_parameters.header)
        mocker.patch("otx.algorithms.detection.tasks.openvino.OpenvinoAdapter")
        mocked_model = mocker.patch.object(Model, "create_model")
        mocked_model.return_value = mocker.MagicMock(spec=Model)
        ov_inferencer = OpenVINODetectionInferencer(params, label_schema, "")
        ov_inferencer.model.__model__ = "OTX_SSD"
        task_env.model = otx_model
        mocker.patch.object(OpenVINODetectionTask, "load_inferencer", return_value=ov_inferencer)
        mocker.patch.object(OpenVINODetectionTask, "load_config", return_value={})

        self.ov_task = OpenVINODetectionTask(task_env)

    @e2e_pytest_unit
    def test_infer(self, mocker):
        """Test infer method in OpenVINODetectionTask."""
        self.dataset, labels = generate_det_dataset(task_type=TaskType.DETECTION)
        fake_ann_scene = self.dataset[0].annotation_scene
        mock_predict = mocker.patch.object(
            OpenVINODetectionInferencer, "predict", return_value=(fake_ann_scene, (None, None))
        )
        updated_dataset = self.ov_task.infer(self.dataset)

        mock_predict.assert_called()
        for updated in updated_dataset:
            assert updated.annotation_scene.contains_any([LabelEntity(name=labels[0].name, domain="DETECTION")])

    @e2e_pytest_unit
    def test_evaluate(self, mocker):
        """Test evaluate method in OpenVINODetectionTask."""
        result_set = ResultSetEntity(
            model=None,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.f_measure.FMeasure", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics="mAP"
        )
        mocker.patch.object(MetricsHelper, "compute_f_measure", return_value=fake_metrics)
        self.ov_task.evaluate(result_set)

        assert result_set.performance.score.value == 0.1

    @e2e_pytest_unit
    def test_deploy(self, otx_model):
        """Test deploy method in OpenVINODetectionTask."""
        output_model = copy.deepcopy(otx_model)
        self.ov_task.model.set_data("openvino.bin", b"foo")
        self.ov_task.model.set_data("openvino.xml", b"bar")
        self.ov_task.config = {"tiling_parameters": None}
        self.ov_task.deploy(output_model)

        assert output_model.exportable_code is not None

    @e2e_pytest_unit
    def test_optimize(self, mocker, otx_model):
        """Test optimize method in OpenVINODetectionTask."""

        def patch_save_model(model, dir_path, model_name):
            with open(f"{dir_path}/{model_name}.xml", "wb") as f:
                f.write(b"foo")
            with open(f"{dir_path}/{model_name}.bin", "wb") as f:
                f.write(b"bar")

        dataset, _ = generate_det_dataset(task_type=TaskType.DETECTION)
        output_model = copy.deepcopy(otx_model)
        self.ov_task.model.set_data("openvino.bin", b"foo")
        self.ov_task.model.set_data("openvino.xml", b"bar")
        mocker.patch("otx.algorithms.detection.tasks.openvino.load_model", autospec=True)
        mocker.patch("otx.algorithms.detection.tasks.openvino.create_pipeline", autospec=True)
        mocker.patch("otx.algorithms.detection.tasks.openvino.save_model", new=patch_save_model)
        spy_compress = mocker.spy(otx.algorithms.detection.tasks.openvino, "compress_model_weights")
        self.ov_task.optimize(OptimizationType.POT, dataset=dataset, output_model=output_model)

        spy_compress.assert_called_once()
        assert self.ov_task.model.get_data("openvino.bin")
        assert self.ov_task.model.get_data("openvino.xml")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.registry import STAGES
from otx.mpa.seg.exporter import SegExporter
from otx.mpa.utils.logger import get_logger

from .stage import SemiSLSegStage

logger = get_logger()


@STAGES.register_module()
class SemiSLSegExporter(SemiSLSegStage, SegExporter):
    def __init__(self, **kwargs):
        SemiSLSegStage.__init__(self, **kwargs)

    def configure(self, model_cfg, model_ckpt, data_cfg, training=False, **kwargs):
        cfg = SemiSLSegStage.configure(self, model_cfg, model_ckpt, data_cfg, training=training, **kwargs)

        cfg.model.type = cfg.model.orig_type
        cfg.model.pop("orig_type", False)
        cfg.model.pop("unsup_weight", False)
        cfg.model.pop("semisl_start_iter", False)

        return cfg



"""Model Configuration of VFNet model for Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = ["../../../../common/adapters/mmcv/configs/backbones/resnet50.yaml", "../../base/models/detector.py"]

model = dict(
    type="CustomVFNet",
    neck=dict(
        type="FPN",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        start_level=1,
        add_extra_convs="on_output",
        num_outs=5,
        relu_before_extra_convs=True,
    ),
    bbox_head=dict(
        type="CustomVFNetHead",
        num_classes=2,
        in_channels=256,
        stacked_convs=3,
        feat_channels=256,
        strides=[8, 16, 32, 64, 128],
        center_sampling=False,
        dcn_on_last_conv=False,
        use_atss=True,
        use_vfl=True,
        loss_cls=dict(
            type="VarifocalLoss",
            use_sigmoid=True,
            alpha=0.75,
            gamma=2.0,
            iou_weighted=True,
            loss_weight=1.0,
        ),
        loss_bbox=dict(type="GIoULoss", loss_weight=1.5),
        loss_bbox_refine=dict(type="GIoULoss", loss_weight=2.0),
    ),
    train_cfg=dict(
        assigner=dict(type="ATSSAssigner", topk=9),
        allowed_border=-1,
        pos_weight=-1,
        debug=False,
    ),
    test_cfg=dict(
        nms_pre=1000,
        min_bbox_size=0,
        score_thr=0.01,
        nms=dict(type="nms", iou_threshold=0.5),
        max_per_img=100,
    ),
)

load_from = "https://storage.openvinotoolkit.org/repositories/openvino_training_extensions\
/models/object_detection/v2/resnet50-vfnet.pth"

__img_norm_cfg = dict(
    mean=[123.675, 116.28, 103.53],
    std=[58.395, 57.12, 57.375],
    to_rgb=True,
)

data = dict(
    pipeline_options=dict(
        MinIouRandomCrop=dict(min_crop_size=0.1),
        Resize=dict(
            img_scale=[(1344, 480), (1344, 960)],
            multiscale_mode="range",
        ),
        Normalize=dict(**__img_norm_cfg),
        MultiScaleFlipAug=dict(
            img_scale=(1344, 800),
            flip=False,
            transforms=[
                dict(type="Resize", keep_ratio=False),
                dict(type="Normalize", **__img_norm_cfg),
                dict(type="Pad", size_divisor=32),
                dict(type="ImageToTensor", keys=["img"]),
                dict(type="Collect", keys=["img"]),
            ],
        ),
    ),
)



"""OTX Algorithms - Classification Dataset."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .datasets import (
    OTXClsDataset,
    OTXHierarchicalClsDataset,
    OTXMultilabelClsDataset,
    SelfSLDataset,
)
from .pipelines import (
    GaussianBlur,
    LoadImageFromOTXDataset,
    OTXColorJitter,
    PILImageToNDArray,
    PostAug,
    RandomAppliedTrans,
)

__all__ = [
    "OTXClsDataset",
    "OTXMultilabelClsDataset",
    "OTXHierarchicalClsDataset",
    "SelfSLDataset",
    "PostAug",
    "PILImageToNDArray",
    "LoadImageFromOTXDataset",
    "RandomAppliedTrans",
    "GaussianBlur",
    "OTXColorJitter",
]



import numpy as np
import pytest
from PIL import Image

from otx.algorithms.classification.adapters.mmcls.data.pipelines import (
    GaussianBlur,
    LoadImageFromOTXDataset,
    OTXColorJitter,
    PILImageToNDArray,
    PostAug,
    RandomAppliedTrans,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit

from .test_datasets import create_cls_dataset


@pytest.fixture(scope="module")
def inputs_np():
    return {"img": np.random.randint(0, 10, (16, 16, 3), dtype=np.uint8), "img_fields": ["img"]}


@pytest.fixture(scope="module")
def inputs_PIL():
    return {
        "img": Image.fromarray(np.random.randint(0, 10, (16, 16, 3), dtype=np.uint8)),
    }


@e2e_pytest_unit
@pytest.mark.parametrize("to_float32", [False, True])
def test_load_image_from_otx_dataset_call(to_float32):
    """Test LoadImageFromOTXDataset."""
    otx_dataset, labels = create_cls_dataset()
    load_image_from_otx_dataset = LoadImageFromOTXDataset(to_float32)
    results = dict(
        dataset_item=otx_dataset[0],
        width=otx_dataset[0].width,
        height=otx_dataset[0].height,
        index=0,
        ann_info=dict(label_list=labels),
    )

    results = load_image_from_otx_dataset(results)

    assert "filename" in results
    assert "ori_filename" in results
    assert "img" in results
    assert "img_shape" in results
    assert "ori_shape" in results
    assert "pad_shape" in results
    assert "img_norm_cfg" in results
    assert "img_fields" in results
    assert isinstance(results["img"], np.ndarray)


@e2e_pytest_unit
def test_random_applied_transforms(mocker, inputs_np):
    """Test RandomAppliedTrans."""
    mocker.patch("otx.algorithms.classification.adapters.mmcls.data.pipelines.build_from_cfg", return_value=lambda x: x)

    random_applied_transforms = RandomAppliedTrans(transforms=[dict()])

    results = random_applied_transforms(inputs_np)

    assert isinstance(results, dict)
    assert "img" in results
    assert repr(random_applied_transforms) == "RandomAppliedTrans"


@e2e_pytest_unit
def test_otx_color_jitter(inputs_np):
    """Test OTXColorJitter."""
    otx_color_jitter = OTXColorJitter()

    results = otx_color_jitter(inputs_np)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_gaussian_blur(inputs_np):
    """Test GaussianBlur."""
    gaussian_blur = GaussianBlur(sigma_min=0.1, sigma_max=0.2)

    results = gaussian_blur(inputs_np)

    assert isinstance(results, dict)
    assert "img" in results
    assert repr(gaussian_blur) == "GaussianBlur"


@e2e_pytest_unit
def test_pil_image_to_nd_array(inputs_PIL) -> None:
    """Test PILImageToNDArray."""
    pil_image_to_nd_array = PILImageToNDArray(keys=["img"])

    results = pil_image_to_nd_array(inputs_PIL)

    assert "img" in results
    assert isinstance(results["img"], np.ndarray)
    assert repr(pil_image_to_nd_array) == "PILImageToNDArray"


@e2e_pytest_unit
def test_post_aug(mocker, inputs_np):
    """Test PostAug."""
    mocker.patch("otx.algorithms.classification.adapters.mmcls.data.pipelines.Compose", return_value=lambda x: x)

    post_aug = PostAug(keys=dict(orig=lambda x: x))

    results = post_aug(inputs_np)

    assert isinstance(results, dict)
    assert "img" in results and "img" in results["img_fields"]
    assert "orig" in results and "orig" in results["img_fields"]
    assert repr(post_aug) == "PostAug"



"""This module contains the mapper for ID entities."""
#
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from otx.api.entities.id import ID


class IDMapper:
    """This class maps an `ID` entity to a string, and vice versa."""

    @staticmethod
    def forward(instance: ID) -> str:
        """Serializes ID to str."""

        return str(instance)

    @staticmethod
    def backward(instance: str) -> ID:
        """Deserializes ID from str."""

        return ID(str(instance))



"""MMCV general build functions."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Callable

from mmcv import Config
from torch.utils.data import DataLoader, Dataset

from otx.api.utils.argument_checks import check_input_parameters_type

# pylint: disable-next=unused-import
from ._builder_build_data_parallel import build_data_parallel  # noqa: F401


@check_input_parameters_type()
def build_dataset(
    config: Config,
    subset: str,
    dataset_builder: Callable,
    *,
    consume: bool = False,
) -> Dataset:
    """Build dataset."""

    if subset in ["test", "val"]:
        default_args = dict(test_mode=True)
    else:
        default_args = dict(test_mode=False)

    dataset_cfg = config.data.pop(subset) if consume else config.data.get(subset)
    dataset = dataset_builder(dataset_cfg, default_args)
    return dataset


@check_input_parameters_type()
def build_dataloader(
    dataset,
    config: Config,
    subset: str,
    dataloader_builder: Callable,
    *,
    distributed: bool = False,
    consume: bool = False,
    **kwargs,
) -> DataLoader:
    """Build dataloader."""

    loader_cfg = dict(
        samples_per_gpu=config.data.get("samples_per_gpu", 1),
        workers_per_gpu=config.data.get("workers_per_gpu", 0),
        num_gpus=len(config.gpu_ids),
        dist=distributed,
        seed=config.get("seed", None),
        shuffle=False if subset in ["test", "val"] else True,  # pylint: disable=simplifiable-if-expression
    )

    # The overall dataloader settings
    loader_cfg.update(
        {
            k: v
            for k, v in config.data.items()
            if k
            not in [
                "train",
                "val",
                "test",
                "unlabeled",
                "train_dataloader",
                "val_dataloader",
                "test_dataloader",
                "unlabeled_dataloader",
            ]
        }
    )

    specific_loader_cfg = (
        config.data.pop(f"{subset}_dataloader", {}) if consume else config.data.get(f"{subset}_dataloader", {})
    )
    loader_cfg = {**loader_cfg, **specific_loader_cfg, **kwargs}

    dataloader = dataloader_builder(
        dataset,
        **loader_cfg,
    )
    return dataloader



"""Model configuration of OCR-Lite-HRnet-18 model for Self-SL Segmentation Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = [
    "../model.py",
]

model = dict(
    type="DetConB",
    pretrained="https://storage.openvinotoolkit.org/repositories/openvino_training_extensions\
        /models/custom_semantic_segmentation/litehrnet18_imagenet1k_rsc.pth",
    num_classes=256,
    num_samples=16,
    downsample=4,
    input_transform="resize_concat",
    in_index=[0, 1, 2, 3],
    neck=dict(
        type="SelfSLMLP",
        in_channels=600,
        hid_channels=256,
        out_channels=128,
        norm_cfg=dict(type="BN1d", requires_grad=True),
        with_avg_pool=False,
    ),
    head=dict(
        type="SelfSLMLP",
        in_channels=128,
        hid_channels=256,
        out_channels=128,
        norm_cfg=dict(type="BN1d", requires_grad=True),
        with_avg_pool=False,
    ),
    loss_cfg=dict(type="DetConLoss", temperature=0.1),
)

load_from = None

resume_from = None

fp16 = None



"""Initialization of executors."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .asynchronous import AsyncExecutor
from .sync_pipeline import ChainExecutor
from .synchronous import SyncExecutor

__all__ = [
    "SyncExecutor",
    "AsyncExecutor",
    "ChainExecutor",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
from mmcv.utils import build_from_cfg
from mmdet.datasets import PIPELINES

# import torchvision.transforms.functional as F
from mmdet.datasets.pipelines.formatting import ImageToTensor, to_tensor

# from mmdet.datasets.pipelines.transforms import Normalize
from PIL import Image, ImageFilter

# import cv2 as cv
from torchvision import transforms as T


@PIPELINES.register_module()
class ColorJitter(T.ColorJitter):
    """MMDet adapter"""

    def __init__(self, key_maps=[("img", "img")], **kwargs):
        super().__init__(**kwargs)
        self.key_maps = key_maps

    def forward(self, inputs):
        outputs = inputs.copy()
        for key_map in self.key_maps:
            outputs[key_map[0]] = super().forward(inputs[key_map[1]])
        return outputs


@PIPELINES.register_module()
class RandomGrayscale(T.RandomGrayscale):
    """MMDet adapter"""

    def __init__(self, key_maps=[("img", "img")], **kwargs):
        super().__init__(**kwargs)
        self.key_maps = key_maps

    def forward(self, inputs):
        outputs = inputs.copy()
        for key_map in self.key_maps:
            outputs[key_map[0]] = super().forward(inputs[key_map[1]])
        return outputs


@PIPELINES.register_module()
class RandomErasing(T.RandomErasing):
    """MMDet adapter"""

    def __init__(self, key_maps=[("img", "img")], **kwargs):
        super().__init__(**kwargs)
        self.key_maps = key_maps

    def forward(self, inputs):
        outputs = inputs.copy()
        for key_map in self.key_maps:
            outputs[key_map[0]] = super().forward(inputs[key_map[1]])
        return outputs


@PIPELINES.register_module()
class RandomGaussianBlur(object):
    """Gaussian blur augmentation in SimCLR https://arxiv.org/abs/2002.05709."""

    def __init__(self, sigma_min, sigma_max, key_maps=[("img", "img")]):
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max
        self.key_maps = key_maps

    def __call__(self, inputs):
        outputs = inputs.copy()
        sigma = np.random.uniform(self.sigma_min, self.sigma_max)
        # ksize = 2*int(np.ceil(2.0*sigma)) + 1
        for key_map in self.key_maps:
            img = inputs[key_map[0]]
            img = img.filter(ImageFilter.GaussianBlur(radius=sigma))
            # img = cv.GaussianBlur(img, ksize=(0,0), sigmaX=sigma)
            # img = F.gaussian_blur(img, ksize, [sigma, sigma])
            outputs[key_map[1]] = img
        return outputs

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class RandomApply(T.RandomApply):
    """MMDet adapter"""

    def __init__(self, transform_cfgs, p=0.5):
        transforms = []
        for transform_cfg in transform_cfgs:
            transforms.append(build_from_cfg(transform_cfg, PIPELINES))
        super().__init__(transforms, p=p)


@PIPELINES.register_module()
class NDArrayToTensor(ImageToTensor):
    """MMDet adapter"""

    def __call__(self, results):
        for key in self.keys:
            img = results[key]
            if len(img.shape) < 3:
                img = np.expand_dims(img, -1)
            img = np.ascontiguousarray(img.transpose(2, 0, 1))
            results[key] = to_tensor(img)
        return results


@PIPELINES.register_module()
class NDArrayToPILImage(object):
    def __init__(self, keys=["img"]):
        self.keys = keys

    def __call__(self, results):
        for key in self.keys:
            img = results[key]
            img = Image.fromarray(img, mode="RGB")
            results[key] = img
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class PILImageToNDArray(object):
    def __init__(self, keys=["img"]):
        self.keys = keys

    def __call__(self, results):
        for key in self.keys:
            img = results[key]
            img = np.asarray(img)
            results[key] = img
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class BranchImage(object):
    def __init__(self, key_map={}):
        self.key_map = key_map

    def __call__(self, results):
        for k1, k2 in self.key_map.items():
            if k1 in results:
                results[k2] = results[k1]
            if k1 in results["img_fields"]:
                results["img_fields"].append(k2)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        return repr_str



"""Exportable code for Anomaly tasks."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .anomaly_classification import AnomalyClassification
from .anomaly_detection import AnomalyDetection
from .anomaly_segmentation import AnomalySegmentation
from .base import AnomalyBase

__all__ = ["AnomalyBase", "AnomalyClassification", "AnomalyDetection", "AnomalySegmentation"]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.detection.utils.utils import ColorPalette, generate_label_schema
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class TestColorPaletteInputParamsValidation:
    @staticmethod
    def color_palette():
        return ColorPalette(1)

    @e2e_pytest_unit
    def test_color_palette_init_params_validation(self):
        """
        <b>Description:</b>
        Check ColorPalette object initialization parameters validation

        <b>Input data:</b>
        ColorPalette object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        ColorPalette object initialization parameter
        """
        correct_values_dict = {
            "n": 1,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "n" parameter
            ("n", unexpected_str),
            # Unexpected string is specified as "rng" parameter
            ("rng", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=ColorPalette,
        )

    @e2e_pytest_unit
    def test_color_palette_get_item_params_validation(self):
        """
        <b>Description:</b>
        Check ColorPalette object "__getitem__" method input parameters validation

        <b>Input data:</b>
        ColorPalette object, "n" non-integer object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__getitem__" method
        """
        color_palette = self.color_palette()
        with pytest.raises(ValueError):
            color_palette.__getitem__("unexpected string")  # type: ignore


class TestOTXUtilsFunctionsInputParamsValidation:
    @e2e_pytest_unit
    def test_generate_label_schema_input_params_validation(self):
        """
        <b>Description:</b>
        Check "generate_label_schema" function input parameters validation

        <b>Input data:</b>
        "generate_label_schema" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "generate_label_schema" function
        """
        correct_values_dict = {
            "label_names": ["label_1", "label_2"],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "label_names" parameter
            ("label_names", unexpected_int),
            # Unexpected integer is specified as nested label name
            ("label_names", ["label_1", unexpected_int]),
            # Unexpected integer is specified as "label_domain" parameter
            ("label_domain", unexpected_int),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=generate_label_schema,
        )



"""UI rules configuration."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from .rules import NullUIRules, Rule, UIRules
from .types import Action, Operator

__all__ = [
    "NullUIRules",
    "Rule",
    "UIRules",
    "Action",
    "Operator",
]



"""This module contains the base class for non-deep learning image-based tasks."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc

from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask


class ImageComputerVisionTask(IInferenceTask, metaclass=abc.ABCMeta):
    """A base class for a non-deep learning image-based tasks, which can only perform inference.

    This class inherits from ITask and IReporting.

    Example:
        A cropping task

    >>> class CroppingTask(ImageComputerVisionTask):
    ...     pass
    """



"""Unit Tests for the MPA Dataset Pipeline - Compose."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from unittest.mock import MagicMock

import numpy as np
import pytest
from mmseg.datasets.builder import PIPELINES
from mmseg.datasets.pipelines import RandomCrop

from otx.mpa.modules.datasets.pipelines.compose import MaskCompose, ProbCompose


class TestProbCompose:
    """ProbCompose Unit Tests."""

    def test_inputs(self) -> None:
        """Test when all inputs are correct."""
        transforms = [MagicMock(), MagicMock(), MagicMock()]
        probs = [0.3, 0.4, 0.3]
        prob_compose = ProbCompose(transforms, probs)
        assert prob_compose.transforms == transforms
        assert np.array_equal(prob_compose.limits, np.array([0.0, 0.3, 0.7, 1.0]))

        # Test when transforms and probs have different lengths
        transforms = [MagicMock(), MagicMock()]
        probs = [0.5, 0.5, 0.5]
        with pytest.raises(AssertionError):
            ProbCompose(transforms, probs)

        # Test when probs have negative values
        transforms = [MagicMock(), MagicMock(), MagicMock()]
        probs = [0.5, -0.2, 0.7]
        with pytest.raises(AssertionError):
            ProbCompose(transforms, probs)

        # Test when sum of probs is 0
        transforms = [MagicMock(), MagicMock(), MagicMock()]
        probs = [0.0, 0.0, 0.0]
        with pytest.raises(AssertionError):
            ProbCompose(transforms, probs)

    def test_dict_transforms(self) -> None:
        """Test whether dict transforms are correctly appended."""
        prob_compose = ProbCompose(transforms=[dict(type="Resize")], probs=[0.7])
        assert (
            repr(prob_compose.transforms[0])
            == "Resize(img_scale=None, multiscale_mode=range, ratio_range=None, keep_ratio=True)"
        )

    def test_invalid_transform_type(self) -> None:
        """Test invalid transform type raises error."""
        with pytest.raises(TypeError):
            transforms = ["Dummy Transform"]
            probs = [0.5]
            pipeline = ProbCompose(transforms, probs)
            del pipeline  # This is to silence the pylint's unused variable warning.

    def test_repr(self) -> None:
        """Test the repr method."""
        transforms = [MagicMock()]
        probs = [0.3]
        prob_compose = ProbCompose(transforms, probs)
        expected_repr = f"ProbCompose(\n    {transforms[0]}\n)"
        assert repr(prob_compose) == expected_repr


@PIPELINES.register_module()
class TestTransform(object):
    def __init__(self) -> None:
        pass

    def __call__(self, data: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
        return data


@pytest.fixture()
def data() -> dict[str, np.ndarray]:
    return {"img": np.ones((100, 100, 3), dtype=np.uint8), "label": np.ones((100, 100))}


class TestMaskCompose:
    """MaskCompose Unit Tests."""

    def test_keep_original_false(self, data: dict[str, np.ndarray]) -> None:
        """Test that the main image is replaced with a mixed image when keep_original is False."""
        transforms = [dict(type="TestTransform")]
        pipeline = MaskCompose(transforms=transforms, prob=1.0, keep_original=False)
        mixed_data = pipeline(data)

        assert np.array_equal(data["img"], mixed_data["img"])
        assert np.array_equal(data["label"], mixed_data["label"])

    def test_keep_original_true(self, data: dict[str, np.ndarray]) -> None:
        """Test that the mixed image is added as aux_img when keep_original is True."""
        transforms = [dict(type="TestTransform")]
        pipeline = MaskCompose(transforms=transforms, prob=1.0, keep_original=True)
        mixed_data = pipeline(data)

        assert np.array_equal(mixed_data["img"], mixed_data["aux_img"])
        assert np.array_equal(data["img"], mixed_data["img"])
        assert np.array_equal(data["label"], mixed_data["label"])

    def test_callable_transform(self, data: dict[str, np.ndarray]) -> None:
        """Test callable transform is appended to the list of transforms."""
        crop_size = (10, 10)
        transform = RandomCrop(crop_size=crop_size)
        pipeline = MaskCompose(transforms=[transform], prob=1.0, keep_original=False)
        mixed_data = pipeline(data)
        assert mixed_data["img_shape"][:2] == crop_size

    def test_invalid_transform_type(self) -> None:
        """Test invalid transform type raises error."""
        with pytest.raises(TypeError):
            transform = [{"Dummy Transform": "So dummy"}]
            pipeline = MaskCompose(transforms=[transform], prob=1.0, keep_original=False)
            del pipeline  # This is to silence the pylint's unused variable warning.

    def test_prob_zero(self, data: dict[str, np.ndarray]) -> None:
        """Test that the main image is not modified when prob is 0."""
        transforms = [
            dict(type="RandomFlip", prob=0.0, direction="horizontal"),
            dict(type="RandomRotate", prob=0.0, degree=30, pad_val=0, seg_pad_val=255),
        ]
        pipeline = MaskCompose(transforms=transforms, prob=0, keep_original=False)
        mixed_data = pipeline(data)

        assert np.array_equal(data["img"], mixed_data["img"])
        assert np.array_equal(data["label"], mixed_data["label"])

    def test_apply_transforms_returns_none(self, data: dict[str, np.ndarray]) -> None:
        """Test that None is returned when apply_transforms returns None."""
        transforms = [
            dict(type="RandomFlip", prob=0.5, direction="horizontal"),
            lambda x: None,
        ]
        pipeline = MaskCompose(transforms=transforms, prob=1.0, keep_original=False)

        with pytest.raises(AssertionError):
            mixed_data = pipeline(data)
            del mixed_data  # This is to silence the pylint's unused variable warning.

    def test_repr(self) -> None:
        """Test __repr__ method."""
        pipeline = MaskCompose(transforms=[dict(type="RandomFlip")], prob=1.0)
        expected_repr = "MaskCompose(\n    RandomFlip(prob=None)\n)"
        assert repr(pipeline) == expected_repr



"""This file defines the ModelConfiguration, ModelEntity and Model classes."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import datetime
from enum import IntEnum, auto
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from otx.api.configuration import ConfigurableParameters
from otx.api.entities.id import ID
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.metrics import NullPerformance, Performance
from otx.api.entities.model_template import TargetDevice
from otx.api.entities.url import URL
from otx.api.usecases.adapters.model_adapter import (
    ExportableCodeAdapter,
    IDataSource,
    ModelAdapter,
)
from otx.api.utils.time_utils import now

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports
    from otx.api.entities.datasets import DatasetEntity


class ModelPrecision(IntEnum):
    """Represents the ModelPrecision of a Model."""

    INT4 = auto()
    INT8 = auto()
    FP16 = auto()
    FP32 = auto()

    def __str__(self):
        """String."""
        return self.name


class ModelConfiguration:
    """This class represents the task configuration which was used to generate a specific model.

    Those are the parameters that a task may need in order to use the model.

    Args:
        configurable_parameters: Task configurable parameters used to
            generate the model
        label_schema: Label schema inside the project used to generate
            the model
    """

    configurable_parameters: ConfigurableParameters

    def __init__(
        self,
        configurable_parameters: ConfigurableParameters,
        label_schema: LabelSchemaEntity,
    ):
        self.configurable_parameters = configurable_parameters
        self.__label_schema = label_schema

    def get_label_schema(self) -> LabelSchemaEntity:
        """Get the LabelSchema."""
        return self.__label_schema


class ModelFormat(IntEnum):
    """Indicate the format of the model."""

    OPENVINO = auto()
    BASE_FRAMEWORK = auto()
    ONNX = auto()


class ModelOptimizationType(IntEnum):
    """Represents optimization type that is used to optimize the model."""

    NONE = auto()
    MO = auto()
    NNCF = auto()
    POT = auto()


class OptimizationMethod(IntEnum):
    """Represents optimization method that is used to optimize the model."""

    FILTER_PRUNING = auto()
    QUANTIZATION = auto()


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class ModelEntity:
    """Represents the Entity of a Model."""

    # TODO: add tags and allow filtering on those in modelrepo
    # pylint: disable=too-many-arguments,too-many-locals; Requires refactor
    def __init__(
        self,
        train_dataset: "DatasetEntity",
        configuration: ModelConfiguration,
        *,
        creation_date: Optional[datetime.datetime] = None,
        performance: Optional[Performance] = None,
        previous_trained_revision: Optional["ModelEntity"] = None,
        previous_revision: Optional["ModelEntity"] = None,
        version: int = 1,
        tags: Optional[List[str]] = None,
        model_format: ModelFormat = ModelFormat.OPENVINO,
        training_duration: float = 0.0,
        model_adapters: Optional[Dict[str, ModelAdapter]] = None,
        exportable_code_adapter: Optional[ExportableCodeAdapter] = None,
        precision: Optional[List[ModelPrecision]] = None,
        latency: int = 0,
        fps_throughput: int = 0,
        target_device: TargetDevice = TargetDevice.CPU,
        target_device_type: Optional[str] = None,
        optimization_type: ModelOptimizationType = ModelOptimizationType.NONE,
        optimization_methods: List[OptimizationMethod] = None,
        optimization_objectives: Dict[str, str] = None,
        performance_improvement: Dict[str, float] = None,
        model_size_reduction: float = 0.0,
        _id: Optional[ID] = None,
    ):
        _id = ID() if _id is None else _id
        performance = NullPerformance() if performance is None else performance
        creation_date = now() if creation_date is None else creation_date

        optimization_methods = [] if optimization_methods is None else optimization_methods
        optimization_objectives = {} if optimization_objectives is None else optimization_objectives
        performance_improvement = {} if performance_improvement is None else performance_improvement

        tags = [] if tags is None else tags
        precision = [ModelPrecision.FP32] if precision is None else precision

        if model_adapters is None:
            model_adapters = {}

        self.__id_ = _id
        self.__creation_date = creation_date
        self.__train_dataset = train_dataset
        self.__previous_trained_revision = previous_trained_revision
        self.__previous_revision = previous_revision
        self.__version = version
        self.__tags = tags
        self.__model_format = model_format
        self.__performance = performance
        self.__training_duration = training_duration
        self.__configuration = configuration
        self.__model_adapters = model_adapters
        self.__exportable_code_adapter = exportable_code_adapter
        self.model_adapters_to_delete: List[ModelAdapter] = []
        self.__precision = precision
        self.__latency = latency
        self.__fps_throughput = fps_throughput
        self.__target_device = target_device
        self.__target_device_type = target_device_type
        self.__optimization_type = optimization_type
        self.__optimization_methods = optimization_methods
        self.__optimization_objectives = optimization_objectives
        self.__performance_improvement = performance_improvement
        self.__model_size_reduction = model_size_reduction

    @property
    def id_(self) -> ID:
        """Gets or sets the id of a Model."""
        return self.__id_

    @id_.setter
    def id_(self, value: ID):
        self.__id_ = value

    @property
    def id(self) -> ID:
        """DEPRECATED."""
        return self.__id_

    @id.setter
    def id(self, value: ID):
        """DEPRECATED."""
        self.__id_ = value

    @property
    def configuration(self) -> ModelConfiguration:
        """Gets or sets the configuration of the Model."""
        return self.__configuration

    @configuration.setter
    def configuration(self, value: ModelConfiguration):
        self.__configuration = value

    @property
    def creation_date(self) -> datetime.datetime:
        """Gets or sets the creation_date of the Model."""
        return self.__creation_date

    @creation_date.setter
    def creation_date(self, value: datetime.datetime):
        self.__creation_date = value

    @property
    def train_dataset(self) -> "DatasetEntity":
        """Gets or sets the current Training Dataset."""
        return self.__train_dataset

    @train_dataset.setter
    def train_dataset(self, value: "DatasetEntity"):
        self.__train_dataset = value

    @property
    def previous_trained_revision(self) -> Union[None, "ModelEntity"]:
        """Gets or sets the previous model.

        Returns:
            None if no previous_trained_revision has been created
        """
        return self.__previous_trained_revision

    @previous_trained_revision.setter
    def previous_trained_revision(self, value: "ModelEntity"):
        self.__previous_trained_revision = value

    @property
    def previous_revision(self) -> Union[None, "ModelEntity"]:
        """Gets or sets the previous model."""
        return self.__previous_revision

    @previous_revision.setter
    def previous_revision(self, value: "ModelEntity"):
        self.__previous_revision = value

    @property
    def version(self) -> int:
        """Gets or sets the version."""
        return self.__version

    @version.setter
    def version(self, value: int):
        self.__version = value

    @property
    def tags(self) -> List[str]:
        """Gets or sets the tags of the Model."""
        return self.__tags

    @tags.setter
    def tags(self, value: List[str]):
        self.__tags = value

    @property
    def model_format(self) -> ModelFormat:
        """Gets the model format."""
        return self.__model_format

    @model_format.setter
    def model_format(self, value: ModelFormat):
        self.__model_format = value

    @property
    def performance(self) -> Performance:
        """Gets or sets the current Performance of the Model."""
        return self.__performance

    @performance.setter
    def performance(self, value: Performance):
        self.__performance = value

    @property
    def training_duration(self) -> float:
        """Gets or sets the current training duration."""
        return self.__training_duration

    @training_duration.setter
    def training_duration(self, value: float):
        self.__training_duration = value

    @property
    def precision(self) -> List[ModelPrecision]:
        """Get or set the precision for the model.

        This has effect on accuracy, latency and throughput of the model.
        """
        return self.__precision

    @precision.setter
    def precision(self, value: List[ModelPrecision]):
        self.__precision = value

    @property
    def latency(self) -> int:
        """Get or set the latency of the model.

        Unit is milliseconds (ms)
        """
        return self.__latency

    @latency.setter
    def latency(self, value: int):
        self.__latency = value

    @property
    def fps_throughput(self) -> int:
        """Get or set the throughput of the model.

        Unit is frames per second (fps)
        """
        return self.__fps_throughput

    @fps_throughput.setter
    def fps_throughput(self, value: int):
        self.__fps_throughput = value

    @property
    def target_device(self) -> TargetDevice:
        """Get or set the device on which the model will be deployed."""
        return self.__target_device

    @target_device.setter
    def target_device(self, value: TargetDevice):
        self.__target_device = value

    @property
    def target_device_type(self) -> Optional[str]:
        """Get or set the type of the target device used by the model."""
        return self.__target_device_type

    @target_device_type.setter
    def target_device_type(self, value: str):
        self.__target_device_type = value

    @property
    def optimization_methods(self) -> Optional[List[OptimizationMethod]]:
        """Get or set the optimization methods used on the model."""
        return self.__optimization_methods

    @optimization_methods.setter
    def optimization_methods(self, value: List[OptimizationMethod]):
        self.__optimization_methods = value

    @property
    def optimization_type(self) -> ModelOptimizationType:
        """Get or set the optimization type used for the model."""
        return self.__optimization_type

    @optimization_type.setter
    def optimization_type(self, value: ModelOptimizationType):
        self.__optimization_type = value

    @property
    def optimization_objectives(self) -> Optional[Dict[str, str]]:
        """Get or set the optimization level of the model."""
        return self.__optimization_objectives

    @optimization_objectives.setter
    def optimization_objectives(self, value: Dict[str, str]):
        self.__optimization_objectives = value

    @property
    def performance_improvement(self) -> Optional[Dict[str, float]]:
        """Get or set the performance improvement of the model."""
        return self.__performance_improvement

    @performance_improvement.setter
    def performance_improvement(self, value: Dict[str, float]):
        self.__performance_improvement = value

    @property
    def model_size_reduction(self) -> float:
        """Get or set the reduction in model size by optimizing."""
        return self.__model_size_reduction

    @model_size_reduction.setter
    def model_size_reduction(self, value: float):
        self.__model_size_reduction = value

    @property
    def exportable_code(self) -> Optional[bytes]:
        """Get the exportable_code from the exportable code adapter."""
        if self.__exportable_code_adapter is not None:
            return self.__exportable_code_adapter.data
        return None

    @exportable_code.setter
    def exportable_code(self, data: Union[bytes, IDataSource]):
        """Set the exportable code using the exportable code adapter."""
        self.__exportable_code_adapter = ExportableCodeAdapter(data_source=data)

    @property
    def exportable_code_adapter(self) -> Optional[ExportableCodeAdapter]:
        """Returns the exportable code adapter."""
        return self.__exportable_code_adapter

    def get_data(self, key: str) -> bytes:
        """Fetches byte data for a certain model.

        Args:
            key: key to fetch data for

        Returns:
            bytes: data for the key.
        """
        return self.__model_adapters[key].data

    def set_data(self, key: str, data: Union[bytes, IDataSource], skip_deletion=False):
        """Sets the data for a specified key, either from a binary blob or from a data source.

        If the key already exists it appends existing data url to a list of urls that will be removed upon saving the
        model. Skip deletion parameter should only be true if replacing bytes data with a file.
        """
        if not skip_deletion:
            self.delete_data(key)
        self.__model_adapters[key] = ModelAdapter(data)

    def delete_data(self, key: str):
        """This function is used to delete data sources that are on the filesystem.

        If the key exists the model adapter will be appended to a list of model adapter that will be removed once the
        model is saved by the repo. Note that an optimized model must contain at least 1 DataSource otherwise you are
        left with an invalid optimized model.
        """
        if key in self.__model_adapters:
            self.model_adapters_to_delete.append(self.__model_adapters[key])
            del self.__model_adapters[key]

    @property
    def model_adapters(self) -> Dict[str, ModelAdapter]:
        """Returns the dictionary of model adapters for each data key."""
        return self.__model_adapters

    @property
    def weight_paths(self) -> Dict[str, URL]:
        """Returns the the path to URLs for each data key.

        Note that this function will raise an error if the model was not saved to a database.
        """
        return {
            key: model_adapter.data_source.binary_url
            for key, model_adapter in self.model_adapters.items()
            if model_adapter.from_file_storage
        }

    def is_optimized(self) -> bool:
        """Returns a boolean indicating if the model has been optimized or not."""
        if self.optimization_type == ModelOptimizationType.NONE:
            return False
        return True

    def __eq__(self, other) -> bool:
        """Compares if both the ModelEntities use the same dataset and have the same performance.

        Args:
            other (ModelEntity): ModelEntity to compare with.

        Returns:
            bool: True if the two ModelEntities are equal, False otherwise.
        """
        if isinstance(other, ModelEntity):
            return (
                self.id_ == other.id_
                and self.train_dataset == other.train_dataset
                and self.performance == other.performance
            )
        return False



"""Async executor based on ModelAPI."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Tuple, Union

import numpy as np
from openvino.model_zoo.model_api.pipelines import AsyncPipeline

from otx.api.usecases.exportable_code.demo.demo_package.model_container import (
    ModelContainer,
)
from otx.api.usecases.exportable_code.demo.demo_package.utils import (
    create_output_converter,
)
from otx.api.usecases.exportable_code.streamer import get_streamer
from otx.api.usecases.exportable_code.visualizers import Visualizer


class AsyncExecutor:
    """Async inferencer.

    Args:
        model: model for inference
        visualizer: visualizer of inference results
    """

    def __init__(self, model: ModelContainer, visualizer: Visualizer) -> None:
        self.model = model.core_model
        self.visualizer = visualizer
        self.converter = create_output_converter(model.task_type, model.labels)
        self.async_pipeline = AsyncPipeline(self.model)

    def run(self, input_stream: Union[int, str], loop: bool = False) -> None:
        """Async inference for input stream (image, video stream, camera)."""
        streamer = get_streamer(input_stream, loop)
        next_frame_id = 0
        next_frame_id_to_show = 0
        stop_visualization = False

        for frame in streamer:
            results = self.async_pipeline.get_result(next_frame_id_to_show)
            while results:
                output = self.render_result(results)
                next_frame_id_to_show += 1
                self.visualizer.show(output)
                if self.visualizer.is_quit():
                    stop_visualization = True
                results = self.async_pipeline.get_result(next_frame_id_to_show)
            if stop_visualization:
                break
            self.async_pipeline.submit_data(frame, next_frame_id, {"frame": frame})
            next_frame_id += 1
        self.async_pipeline.await_all()
        for next_frame_id_to_show in range(next_frame_id_to_show, next_frame_id):
            results = self.async_pipeline.get_result(next_frame_id_to_show)
            output = self.render_result(results)
            self.visualizer.show(output)

    def render_result(self, results: Tuple[Any, dict]) -> np.ndarray:
        """Render for results of inference."""
        predictions, frame_meta = results
        annotation_scene = self.converter.convert_to_annotation(predictions, frame_meta)
        current_frame = frame_meta["frame"]
        output = self.visualizer.draw(current_frame, annotation_scene, frame_meta)
        return output



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path

import pytest
from omegaconf import OmegaConf

from otx.api.configuration import cfg_helper
from otx.api.configuration.elements import metadata_keys
from otx.api.configuration.enums import AutoHPOState, ModelLifecycle
from tests.unit.api.configuration.dummy_config import (
    DatasetManagerConfig,
    SomeEnumSelectable,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestConfigurationHelper:
    @staticmethod
    def __get_path_to_file(filename: str):
        """
        Return the path to the file named 'filename', which lives in the tests/configuration directory
        """
        return str(Path(__file__).parent / Path(filename))

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_config_reconstruction(self):
        """
        <b>Description:</b>
        This test verifies that a configuration can be converted to a dictionary and yaml string, and back again. The
        test passes if the reconstructed configuration is equal to the initial one.

        <b>Input data:</b>
        Dummy configuration -- DatasetManagerConfig

        <b>Expected results:</b>
        Test passes if reconstructed configuration is equal to the original one

        <b>Steps</b>
        1. Create configuration
        2. Convert to dictionary and yaml string representation
        3. Reconstruct configuration from dict
        4. Reconstruct configuration from yaml string
        5. Verify that contents of reconstructed configs are equal to original config
        """
        # Initialize the config object
        config = DatasetManagerConfig(
            description="Configurable parameters for the DatasetManager -- TEST ONLY",
            header="Dataset Manager configuration -- TEST ONLY",
        )

        # Convert config to dictionary and to yaml string
        cfg = cfg_helper.convert(config, dict)
        cfg_yaml = cfg_helper.convert(config, str)

        # Reconstruct the config from dictionary and from yaml string
        reconstructed_config = cfg_helper.create(cfg)
        reconstructed_config_from_yaml = cfg_helper.create(cfg_yaml)

        # Compare the config dictionaries. Order of some parameters may change in the conversion, so dictionary
        # comparison will work while comparing objects or yaml strings directly likely does not result in equality.
        assert cfg == cfg_helper.convert(reconstructed_config, dict)
        assert cfg == cfg_helper.convert(reconstructed_config_from_yaml, dict)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_config_reconstruction_with_metadata_change(self):
        """
        <b>Description:</b>
        This test verifies that metadata changes are capture upon converting a
        configuration to a dictionary and yaml string, and back again. The test passes
        if the reconstructed configuration is equal to the initial one, with
        non-default metadata values set.

        <b>Input data:</b>
        Dummy configuration -- DatasetManagerConfig

        <b>Expected results:</b>
        Test passes if metadata changes are reflected in the reconstructed
        configuration, such that it is equal to the original one

        <b>Steps</b>
        1. Create configuration
        2. Set metadata for a parameter to a non-default value
        3. Convert to dictionary and yaml string representation
        4. Reconstruct configuration from dict
        5. Reconstruct configuration from yaml string
        6. Verify that contents of reconstructed configs are equal to original config
        """
        # Arrange
        # Initialize the config object
        config = DatasetManagerConfig(
            description="Configurable parameters for the DatasetManager -- TEST ONLY",
            header="Dataset Manager configuration -- TEST ONLY",
        )
        test_parameter_name = "dummy_float_selectable"
        metadata_key = metadata_keys.AUTO_HPO_STATE
        old_state = config.get_metadata(test_parameter_name)[metadata_key]
        new_state = AutoHPOState.OPTIMIZED
        set_success = config.set_metadata_value(
            parameter_name=test_parameter_name,
            metadata_key=metadata_key,
            value=new_state,
        )

        # Act
        # Convert config to dictionary and to yaml string
        cfg = cfg_helper.convert(config, dict)
        cfg_yaml = cfg_helper.convert(config, str)

        # Reconstruct the config from dictionary and from yaml string
        reconstructed_config = cfg_helper.create(cfg)
        reconstructed_config_from_yaml = cfg_helper.create(cfg_yaml)

        # Assert
        assert old_state != new_state
        assert set_success
        # Check that metadata changes are properly converted
        assert reconstructed_config.get_metadata(parameter_name=test_parameter_name)[metadata_key] == new_state
        assert (
            reconstructed_config_from_yaml.get_metadata(parameter_name=test_parameter_name)[metadata_key] == new_state
        )
        # Compare the config dictionaries
        assert cfg == cfg_helper.convert(reconstructed_config, dict)
        assert cfg == cfg_helper.convert(reconstructed_config_from_yaml, dict)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_creation_from_yaml(self):
        """
        <b>Description:</b>
        This test verifies that a configuration can be created from a yaml file

        <b>Input data:</b>
        dummy_config.py   -- DatasetManagerConfig class definition
        dummy_config.yaml -- yaml file specifying a configuration equivalent to DatasetManagerConfig, but in yaml

        <b>Expected results:</b>
        Test passes if the contents of the config created from dummy_config.yaml is equal to the configuration in
        DatasetManagerConfig

        <b>Steps</b>
        1. Create configuration from class
        2. Create configuration from yaml
        3. Convert both configs to dictionary
        4. Compare resulting dictionaries
        """
        # Initialize the config
        config = DatasetManagerConfig(
            description="Configurable parameters for the DatasetManager -- TEST ONLY",
            header="Dataset Manager configuration -- TEST ONLY",
        )

        cfg_from_yaml = cfg_helper.create(self.__get_path_to_file("./dummy_config.yaml"))

        # Compare the config dictionaries. Order of some parameters may change in the conversion, so dictionary
        # comparison will work while comparing objects or yaml strings directly likely does not result in equality.
        cfg_dict = cfg_helper.convert(config, dict)
        cfg_from_yaml_dict = cfg_helper.convert(cfg_from_yaml, dict)

        # Check the parameter groups individually, to narrow down any errors more easily
        cfg_subset_parameters = cfg_dict.pop("subset_parameters")
        yamlcfg_subset_parameters = cfg_from_yaml_dict.pop("subset_parameters")
        assert cfg_subset_parameters == yamlcfg_subset_parameters

        cfg_nested_group = cfg_dict.pop("nested_parameter_group")
        yamlcfg_nested_group = cfg_from_yaml_dict.pop("nested_parameter_group")
        assert cfg_nested_group == yamlcfg_nested_group

        assert config.dummy_selectable == cfg_dict["dummy_selectable"]["value"]

        assert cfg_dict == cfg_from_yaml_dict

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_broken_config(self):
        """
        <b>Description:</b>
        This test verifies that a configuration created from a yaml file that contains invalid parameter values will
        raise a ValueError. It also verifies that no ValueError is raised if the config is created with valid values.

        <b>Input data:</b>
        dummy_broken_config.yaml -- yaml file specifying a configuration, holding an out-of-bounds value for the number
                                    of epochs

        <b>Expected results:</b>
        Test passes if a ValueError is raised upon creating the config from yaml, and no ValueError is raised upon
        config creation after correcting the out-of-bounds value.

        <b>Steps</b>
        1. Create configuration from dummy_broken_config.yaml
        2. Assert that ValueError is raised
        3. Load dummy_broken_config.yaml as dictionary and correct invalid epochs value
        4. Create configuration from the corrected dictionary
        5. Assert that epochs value is set in the configuration correctly
        """
        # Loading a config that has an invalid value (-5 for epochs) should raise a ValueError due to runtime
        # input validation upon config creation.
        broken_config_path = self.__get_path_to_file("dummy_broken_config.yaml")
        with pytest.raises(ValueError) as error:
            config = cfg_helper.create(broken_config_path)

        assert "Invalid value set for epochs: -5 is out of bounds." == str(error.value)

        # Test correcting the broken config by first loading it from the yaml file, and then setting epochs to a valid
        # value. Config should now be created correctly. Finally, assert that the value for epochs has been corrected
        dict_config = OmegaConf.load(broken_config_path)
        dict_config.learning_parameters.epochs.value = 10
        config = cfg_helper.create(dict_config)
        assert config.learning_parameters.epochs == 10

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_validation(self):
        """
        <b>Description:</b>
        This test verifies that validating the parameter values in the configuration works. Validation should raise a
        ValueError if any of the config parameters is set to an invalid value.

        <b>Input data:</b>
        dummy_config.py   -- DatasetManagerConfig class definition


        <b>Expected results:</b>
        Test passes if a ValueError is raised upon validation of the configuration, if it contains invalid values for
        any of its parameters.

        <b>Steps</b>
        1. Create DatasetManagerConfig configuration
        2. Set test_proportion to an out of bounds value
        3. Validate config and assert that ValueError is raised
        4. Set test_proportion to a valid value
        5. Validate config and assert that no ValueError is raised
        6. Set dummy_selectable parameter to invalid value
        7. Validate config and assert that ValueError is raised
        8. Correct dummy_selectable and assert that ValueError is not raised upon validation
        """
        # Initialize the config
        config = DatasetManagerConfig(
            description="TEST ONLY",
            header="TEST ONLY",
        )

        # Assert that config passes validation initially
        assert cfg_helper.validate(config)

        # Set invalid test_proportion, and assert that this raises an error upon validation
        with pytest.raises(ValueError):
            config.subset_parameters.test_proportion = 1.1

        # Assert that validation passes again after restoring a value that is within the bounds
        config.subset_parameters.test_proportion = 0.25
        assert cfg_helper.validate(config)

        # Set value that is not one of the options for dummy_selectable, assert that this raises a ValueError
        with pytest.raises(ValueError):
            config.dummy_selectable = "invalid_value"

        # Assert that validation passes again after restoring to a value that is in the options list
        config.dummy_selectable = "option_c"
        assert cfg_helper.validate(config)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_substitute_values(self):
        """
        <b>Description:</b>
        This test verifies that parameter values can be substituted into a configuration, using an input dictionary or
        input configuration.

        <b>Input data:</b>
        dummy_config.py   -- DatasetManagerConfig class definition

        <b>Expected results:</b>
        Test passes if the configuration is successfully converted into a dictionary,
        with keys and values matching the structure of the original configuration

        <b>Steps</b>
        1. Create configuration from dummy_config.yaml
        2. Check that values are set according to their initial definition in dummy_config.yaml
        3. Load dummy_config.yaml as an input dictionary, and change some of its parameter values.
        4. Substitute the values from the dictionary
        5. Check that the parameter values have been changed correctly in the config
        6. Change the parameter values in the input dictionary again
        7. Convert the input dictionary to an input configuration
        8. Substitute the values from the input configuration
        9. Check that the parameter values have been changed correctly in the config
        """
        # Initialize the config from yaml
        config = cfg_helper.create(self.__get_path_to_file("dummy_config.yaml"))

        # Assert that the values are set according to what is specified in the yaml
        assert config.subset_parameters.test_proportion == 0.15
        assert config.number_of_samples_for_auto_train == 5

        # Load the config as a dict from the yaml and change some of the values
        config_dict = OmegaConf.load(self.__get_path_to_file("dummy_config.yaml"))
        config_dict.subset_parameters.test_proportion.value = 0.05
        config_dict.number_of_samples_for_auto_train.value = 50

        # Substitute values from this dict
        cfg_helper.substitute_values(config, value_input=config_dict)

        # Assert that the values are changed in the config, according to what was substituted above
        assert config.subset_parameters.test_proportion == 0.05
        assert config.number_of_samples_for_auto_train == 50

        # Convert config_dict to actual config object, and then use that as input for the value substitution
        config_dict.subset_parameters.test_proportion.value = 0.80
        config_dict.number_of_samples_for_auto_train.value = 500
        reconstructed_config = cfg_helper.create(config_dict)
        cfg_helper.substitute_values(config, value_input=reconstructed_config)

        # Assert that the values are changed in the config, according to what was substituted above
        assert config.subset_parameters.test_proportion == 0.80
        assert config.number_of_samples_for_auto_train == 500

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_substitute_values_for_lifecycle(self):
        """
        <b>Description:</b>
        This test verifies that parameter values can be substituted into a
        configuration, conditional on the phase in the model lifecycle that they affect

        <b>Input data:</b>
        dummy_config.py   -- DatasetManagerConfig class definition

        <b>Expected results:</b>
        Test passes if the configuration values that affect the model life cycle
        specified are updated, whereas others are not

        <b>Steps</b>
        1. Create configuration from dummy_config.yaml
        2. Check that values are set according to their initial definition in
            dummy_config.yaml
        3. Create second config from dummy_config.yaml, and change some of its values
        4. Substitute the values from the modified config
        5. Check that the parameter values have been changed correctly in the config
        6. Check that parameter values for parameters that do not belong to the target
            model lifecycle did not change
        """
        # Initialize the config from yaml
        config = cfg_helper.create(self.__get_path_to_file("dummy_config.yaml"))

        # Assert that the values are set according to what is specified in the yaml
        assert config.subset_parameters.test_proportion == 0.15
        assert config.subset_parameters.get_metadata("test_proportion")["affects_outcome_of"] == ModelLifecycle.TRAINING
        assert config.dummy_float_selectable == 2
        assert config.get_metadata("dummy_float_selectable")["affects_outcome_of"] == ModelLifecycle.NONE
        assert config.dummy_selectable == SomeEnumSelectable.BOGUS_NAME
        assert config.get_metadata("dummy_selectable")["affects_outcome_of"] == ModelLifecycle.INFERENCE

        # Load the config again from the yaml and change some of the values
        config_2 = cfg_helper.create(self.__get_path_to_file("dummy_config.yaml"))
        config_2.subset_parameters.test_proportion = 0.05
        config_2.dummy_float_selectable = 4.0
        config_2.dummy_selectable = SomeEnumSelectable.TEST_NAME1

        cfg_helper.substitute_values_for_lifecycle(config, config_2, model_lifecycle=ModelLifecycle.INFERENCE)

        assert config.subset_parameters.test_proportion == 0.15
        assert config.dummy_selectable == SomeEnumSelectable.TEST_NAME1
        assert config.dummy_float_selectable == 2

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_substitute_values_for_lifecycle_list(self):
        """
        <b>Description:</b>
        This test verifies that parameter values can be substituted into a
        configuration, conditional on the phase in the model lifecycle that they affect

        <b>Input data:</b>
        dummy_config.py   -- DatasetManagerConfig class definition

        <b>Expected results:</b>
        Test passes if the configuration values that affect the model life cycle
        specified are updated, whereas others are not

        <b>Steps</b>
        1. Create configuration from dummy_config.yaml
        2. Check that values are set according to their initial definition in
            dummy_config.yaml
        3. Create second config from dummy_config.yaml, and change some of its values
        4. Substitute the values from the modified config
        5. Check that the parameter values have been changed correctly in the config
        6. Check that parameter values for parameters that do not belong to the target
            model lifecycle did not change
        """
        # Initialize the config from yaml
        config = cfg_helper.create(self.__get_path_to_file("dummy_config.yaml"))

        # Assert that the values are set according to what is specified in the yaml
        assert config.subset_parameters.test_proportion == 0.15
        assert config.subset_parameters.get_metadata("test_proportion")["affects_outcome_of"] == ModelLifecycle.TRAINING
        assert config.dummy_float_selectable == 2
        assert config.get_metadata("dummy_float_selectable")["affects_outcome_of"] == ModelLifecycle.NONE
        assert config.dummy_selectable == SomeEnumSelectable.BOGUS_NAME
        assert config.get_metadata("dummy_selectable")["affects_outcome_of"] == ModelLifecycle.INFERENCE

        # Load the config again from the yaml and change some of the values
        config_2 = cfg_helper.create(self.__get_path_to_file("dummy_config.yaml"))
        config_2.subset_parameters.test_proportion = 0.05
        config_2.dummy_float_selectable = 4.0
        config_2.dummy_selectable = SomeEnumSelectable.TEST_NAME1

        cfg_helper.substitute_values_for_lifecycle(
            config,
            config_2,
            model_lifecycle=[ModelLifecycle.INFERENCE, ModelLifecycle.NONE],
        )

        assert config.subset_parameters.test_proportion == 0.15
        assert config.dummy_selectable == SomeEnumSelectable.TEST_NAME1
        assert config.dummy_float_selectable == 4

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_values_only_conversion(self):
        """
        <b>Description:</b>
        This test verifies that converting a configuration into a dictionary while
        retaining only parameter values (and discarding metadata) works

        <b>Input data:</b>
        dummy_config.yaml -- yaml file specifying a configuration equivalent to
        DatasetManagerConfig, but in yaml

        <b>Expected results:</b>
        Test passes if the parameter values of the configuration are updated according to the input dictionary and
        input configuration from which the values are substituted.

        <b>Steps</b>
        1. Create DatasetManagerConfig configuration
        2. Change a couple of parameters away from their defaults
        3. Convert the configuration to a dictionary, with `values_only=True` to
            discard meta-data
        4. Assert that the resulting dictionary contains the expected values
        """
        config = DatasetManagerConfig()
        config.subset_parameters.test_proportion = 0.3
        config.dummy_selectable = SomeEnumSelectable.OPTION_C

        config_dict = cfg_helper.convert(config, target=dict, values_only=True)

        assert config_dict["subset_parameters"]["test_proportion"] == 0.3
        assert config_dict["dummy_selectable"] == SomeEnumSelectable.OPTION_C
        assert config_dict["number_of_samples_for_auto_train"] == 5
        assert config_dict["nested_parameter_group"]["subgroup_one"]["bogus_parameter_one"] == 42



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch


def get_valid_label_mask_per_batch(img_metas, num_classes):
    valid_label_mask_per_batch = []
    for _, meta in enumerate(img_metas):
        valid_label_mask = torch.Tensor([1 for _ in range(num_classes)])
        if "ignored_labels" in meta and meta["ignored_labels"]:
            valid_label_mask[meta["ignored_labels"]] = 0
        valid_label_mask_per_batch.append(valid_label_mask)
    return valid_label_mask_per_batch



"""Datasets for testing anomaly tasks"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Dict, Tuple

import albumentations as A
import numpy as np
from albumentations.pytorch import ToTensorV2
from bson import ObjectId
from omegaconf import OmegaConf
from pytorch_lightning.core.datamodule import LightningDataModule
from torch.utils.data import DataLoader

from otx.algorithms.anomaly.adapters.anomalib.data.data import (
    OTXAnomalyDataModule,
    OTXAnomalyDataset,
)
from otx.algorithms.anomaly.adapters.anomalib.data.dataset import (
    AnomalyClassificationDataset,
    AnomalyDetectionDataset,
    AnomalySegmentationDataset,
)
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity, DatasetPurpose
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset


def get_shapes_dataset(task_type: TaskType, one_each: bool = False) -> DatasetEntity:
    """Get shapes dataset.

    Args:
        task_type (TaskType): Task type.
        one_each (bool): If this flag is true then it will sample one normal and one abnormal image for each split.
            The training split will have only one normal image. Defaults to False.
    """
    dataset: DatasetEntity
    if task_type == TaskType.ANOMALY_CLASSIFICATION:
        train_subset, test_subset, val_subset = _get_annotations("classification")
        dataset = AnomalyClassificationDataset(train_subset, val_subset, test_subset)
    elif task_type == TaskType.ANOMALY_SEGMENTATION:
        train_subset, test_subset, val_subset = _get_annotations("segmentation")
        dataset = AnomalySegmentationDataset(train_subset, val_subset, test_subset)
    elif task_type == TaskType.ANOMALY_DETECTION:
        train_subset, test_subset, val_subset = _get_annotations("detection")
        dataset = AnomalyDetectionDataset(train_subset, val_subset, test_subset)
    else:
        raise ValueError(f"{task_type} not supported.")

    if one_each:
        train_subset = []
        test_subset = []
        val_subset = []
        for item in dataset._items:
            if item.subset == Subset.TRAINING and len(train_subset) < 1:
                train_subset.append(item)
            # Check if the label is already present in the subset.
            elif item.subset == Subset.TESTING and item.annotation_scene.annotations[0].get_labels()[0].name not in [
                a.annotation_scene.annotations[0].get_labels()[0].name for a in test_subset
            ]:
                test_subset.append(item)
            # Check if the label is already present in the subset.
            elif item.subset == Subset.VALIDATION and item.annotation_scene.annotations[0].get_labels()[0].name not in [
                a.annotation_scene.annotations[0].get_labels()[0].name for a in val_subset
            ]:
                val_subset.append(item)
        dataset._items = train_subset + test_subset + val_subset
    return dataset


class DummyDataset(OTXAnomalyDataset):
    def __init__(self, task_type: TaskType):
        self.normal_label = LabelEntity(id=ID(0), name="Normal", domain=Domain.ANOMALY_CLASSIFICATION)
        self.abnormal_label = LabelEntity(
            id=ID(1),
            name="Anomalous",
            domain=Domain.ANOMALY_CLASSIFICATION,
            is_anomalous=True,
        )
        self.dataset = self.get_mock_dataitems()
        self.task_type = task_type
        self.transform = A.Compose(
            [
                A.Resize(32, 32, always_apply=True),
                ToTensorV2(),
            ]
        )
        self.config = OmegaConf.create({"dataset": {"image_size": [32, 32]}})

    def get_mock_dataitems(self) -> DatasetEntity:
        dataset_items = []

        image_anomalous = Image(np.ones((32, 32, 1)))
        annotations = [
            Annotation(Rectangle.generate_full_box(), labels=[ScoredLabel(label=self.abnormal_label, probability=1.0)])
        ]
        polygon = Polygon(points=[Point(0.0, 0.0), Point(1.0, 1.0)])
        annotations.append(
            Annotation(shape=polygon, labels=[ScoredLabel(self.abnormal_label, probability=1.0)], id=ID(ObjectId()))
        )
        annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)
        dataset_items.append(DatasetItemEntity(media=image_anomalous, annotation_scene=annotation_scene))

        image_normal = Image(np.zeros((32, 32, 1)))
        annotations = [
            Annotation(Rectangle.generate_full_box(), labels=[ScoredLabel(label=self.normal_label, probability=1.0)])
        ]
        annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)
        dataset_items.append(DatasetItemEntity(media=image_normal, annotation_scene=annotation_scene))

        return DatasetEntity(items=dataset_items, purpose=DatasetPurpose.INFERENCE)


class DummyDataModule(LightningDataModule):
    def __init__(self, task_type: TaskType):
        super().__init__()
        self.dataset = DummyDataset(task_type)
        self.labels = [self.dataset.abnormal_label, self.dataset.normal_label]

    def train_dataloader(self) -> DataLoader:
        return DataLoader(self.dataset, pin_memory=True)

    def test_dataloader(self) -> DataLoader:
        return DataLoader(self.dataset, pin_memory=True)

    def val_dataloader(self) -> DataLoader:
        return DataLoader(self.dataset, pin_memory=True)

    def predict_dataloader(self) -> DataLoader:
        return DataLoader(self.dataset, shuffle=False, pin_memory=True)


class ShapesDataModule(OTXAnomalyDataModule):
    """Creates datamodule with shapes dataset.

    Args:
        task_type (TaskType): Task type (classification, detection, segmentation)
    """

    def __init__(self, task_type: TaskType):
        self.config = OmegaConf.create(
            {
                "dataset": {
                    "eval_batch_size": 32,
                    "train_batch_size": 32,
                    "test_batch_size": 32,
                    "num_workers": 2,
                    "image_size": [32, 32],
                    "transform_config": {"train": None},
                }
            }
        )
        # self.dataset = ShapesDataset(get_shapes_dataset(task_type), task_type, self.config)
        self.dataset = get_shapes_dataset(task_type)
        super().__init__(config=self.config, dataset=self.dataset, task_type=task_type)


def _get_annotations(task: str) -> Tuple[Dict, Dict, Dict]:
    ann_file_root = Path("tests", "assets", "anomaly", task)
    data_root = Path("tests", "assets", "anomaly", "shapes")

    train_subset = {"ann_file": str(ann_file_root / "train.json"), "data_root": str(data_root)}
    test_subset = {"ann_file": str(ann_file_root / "test.json"), "data_root": str(data_root)}

    val_subset = {"ann_file": str(ann_file_root / "val.json"), "data_root": str(data_root)}
    return train_subset, test_subset, val_subset
